from hwt.hdl.types.defs import BIT
from hwt.hdl.types.hdlType import default_auto_cast_fn
from hwt.hdl.value import Value


def cast_hbool(self, sigOrVal, toType):
    if toType == BIT:
        if isinstance(sigOrVal, Value):
            v = BIT.getValueCls()(int(sigOrVal.val), BIT, sigOrVal.vldMask,
                                  sigOrVal.updateTime)
            return v
        else:
            return sigOrVal._ternary(BIT.fromPy(1), BIT.fromPy(0))

    return default_auto_cast_fn(self, sigOrVal, toType)



from hwt.serializer.ip_packager.exprSerializer import VivadoTclExpressionSerializer
from hwt.serializer.ip_packager.helpers import mkSpiElm, spi_ns_prefix
from hwt.serializer.ip_packager.otherXmlObjs import Parameter
from hwt.serializer.vhdl.serializer import VhdlSerializer
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


DEFAULT_CLOCK = 100000000


class Type():
    __slots__ = ['name', 'version', 'vendor', 'library']

    #@classmethod
    #def fromElem(cls, elm):
    #    self = cls()
    #    for s in ['name', 'version', 'vendor', 'library']:
    #        setattr(self, s, elm.attrib[spi_ns_prefix + s])
    #    return self

    def asElem(self, elmName):
        e = mkSpiElm(elmName)
        for s in ['name', 'version', 'vendor', 'library']:
            e.attrib[spi_ns_prefix + s] = getattr(self, s)
        return e


class IntfConfig(Type):
    def __init__(self):
        self.parameters = []
        self.map = {}

    def addSimpleParam(self, thisIntf, name, value):
        p = Parameter()
        p.name = name
        p.value.resolve = "immediate"
        p.value.id = "BUSIFPARAM_VALUE." + thisIntf._name.upper() + "." + name.upper()
        p.value.text = value
        self.parameters.append(p)
        return p

    def addWidthParam(self, thisIntf, name, value):
        ctx = VhdlSerializer.getBaseContext()

        def createTmpVar(suggestedName, dtype):
            raise NotImplementedError("Value of generic %s can not be converted do ipcore format (%s)", name, repr(value))
        ctx.createTmpVarFn = createTmpVar

        p = self.addSimpleParam(thisIntf, "ADDR_WIDTH",
                                VivadoTclExpressionSerializer.asHdl(value.staticEval(), ctx))
        if isinstance(value, RtlSignalBase):
            p.value.resolve = "user"

    def postProcess(self, component, entity, allInterfaces, thisIf):
        pass



from hwt.hdl.hdlObject import HdlObject


class Entity(HdlObject):
    """
    Hdl container of hdl configuration and interfaces
    """

    def __init__(self, name):
        self.name = name
        self.origin = None  # creator of this object
        self.generics = []
        self.ports = []
        self.ctx = {}
        self.discovered = False



from copy import copy
import enum
from operator import eq, ne, lt, gt, ge, le

from hwt.bitmask import mask, selectBit, selectBitRange, bitSetTo, \
    setBitRange
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.typeShortcuts import hInt
from hwt.hdl.types.bitValFunctions import bitsCmp__val, bitsCmp, \
    bitsBitOp__val, bitsBitOp, bitsArithOp__val, bitsArithOp, signFix
from hwt.hdl.types.bitVal_bitOpsVldMask import vldMaskForXor, \
    vldMaskForAnd, vldMaskForOr
from hwt.hdl.types.bitVal_opReduce import tryReduceOr, tryReduceAnd, \
    tryReduceXor
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, INT, BIT, SLICE
from hwt.hdl.types.eventCapableVal import EventCapableVal
from hwt.hdl.types.integer import Integer
from hwt.hdl.types.integerVal import IntegerVal
from hwt.hdl.types.slice import Slice
from hwt.hdl.types.sliceUtils import slice_to_SLICE
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import Value, areValues
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc


class BitsVal(EventCapableVal):
    """
    :attention: operator on signals are using value operator functions as well
    """

    def _isFullVld(self):
        return self.vldMask == self._dtype._allMask

    def _convSign__val(self, signed):
        t = self._dtype
        if t.signed == signed:
            return self
        selfSign = t.signed
        v = self.clone()
        w = t._widthVal
        m = t._allMask
        _v = v.val

        if selfSign and not signed:
            if _v < 0:
                v.val = m + _v + 1
        elif not selfSign and signed:
            msbMask = 1 << (w - 1)
            if _v >= msbMask:
                v.val = -_v + msbMask + (m >> 1) - 1
        v._dtype = v._dtype.__class__(w, signed=signed)

        return v

    def _convSign(self, signed):
        """
        Convert signum, no bit manipulation just data are represented
        differently

        :param signed: if True value will be signed,
            if False value will be unsigned,
            if None value will be vector without any sign specification
        """
        if isinstance(self, Value):
            return self._convSign__val(signed)
        else:
            if self._dtype.signed == signed:
                return self
            t = copy(self._dtype)
            t.signed = signed
            if signed is None:
                cnv = AllOps.BitsAsVec
            elif signed:
                cnv = AllOps.BitsAsSigned
            else:
                cnv = AllOps.BitsAsUnsigned

            return Operator.withRes(cnv, [self], t)

    def _signed(self):
        return self._convSign(True)

    def _unsigned(self):
        return self._convSign(False)

    def _vec(self):
        return self._convSign(None)

    @classmethod
    def fromPy(cls, val, typeObj, vldMask=None):
        """
        Construct value from pythonic value (int, bytes, enum.Enum member)
        """
        assert not isinstance(val, Value)
        if val is None:
            vld = 0
            val = 0
            assert vldMask is None or vldMask == 0
        else:
            allMask = typeObj.all_mask()
            w = typeObj.bit_length()
            if isinstance(val, bytes):
                val = int.from_bytes(
                    val, byteorder="little", signed=bool(typeObj.signed))
            else:
                try:
                    val = int(val)
                except TypeError as e:
                    if isinstance(val, enum.Enum):
                        val = int(val.value)
                    else:
                        raise e

            if vldMask is None:
                vld = allMask
            else:
                assert vldMask <= allMask and vldMask >= 0
                vld = vldMask

            if val < 0:
                assert typeObj.signed
                assert signFix(val & allMask, w) == val, (
                    val, signFix(val & allMask, w))
                val = signFix(val & vld, w)
            else:
                if typeObj.signed:
                    msb = 1 << (w - 1)
                    if msb & val:
                        assert val < 0, val

                if val & allMask != val:
                    raise ValueError(
                        "Not enought bits to represent value",
                        val, val & allMask)
                val = val & vld

        return cls(val, typeObj, vld)

    def _concat__val(self, other):
        w = self._dtype.bit_length()
        other_w = other._dtype.bit_length()
        resWidth = w + other_w
        resT = Bits(resWidth)

        v = self.clone()
        v.val = (v.val << other_w) | other.val
        v.vldMask = (v.vldMask << other_w) | other.vldMask
        v.updateTime = max(self.updateTime, other.updateTime)
        v._dtype = resT

        return v

    def _concat(self, other):
        """
        Concatenate this with other to one wider value/signal
        """
        w = self._dtype.bit_length()
        try:
            other_bit_length = other._dtype.bit_length
        except AttributeError:
            raise TypeError("Can not concat bits and", other._dtype)

        other_w = other_bit_length()
        resWidth = w + other_w
        resT = Bits(resWidth)

        if areValues(self, other):
            return self._concat__val(other)
        else:
            w = self._dtype.bit_length()
            other_w = other._dtype.bit_length()
            resWidth = w + other_w
            resT = Bits(resWidth)
            # is instance of signal
            if isinstance(other, InterfaceBase):
                other = other._sig
            if isinstance(other._dtype, Bits):
                if other._dtype.signed is not None:
                    other = other._vec()
            elif other._dtype == BOOL:
                other = other._auto_cast(BIT)
            else:
                raise TypeError(other._dtype)

            if self._dtype.signed is not None:
                self = self._vec()

            return Operator.withRes(AllOps.CONCAT, [self, other], resT)\
                           ._auto_cast(Bits(resWidth,
                                            signed=self._dtype.signed))

    def _getitem__val_int(self, key):
        updateTime = max(self.updateTime, key.updateTime)
        keyVld = key._isFullVld()

        if keyVld:
            val = selectBit(self.val, key.val)
            vld = selectBit(self.vldMask, key.val)
        else:
            val = 0
            vld = 0

        return self.__class__(val, BIT, vld, updateTime=updateTime)

    def _getitem__val_slice(self, key):
        updateTime = max(self.updateTime, key.updateTime)
        assert key._isFullVld()
        size = key._size()

        firstBitNo = key.val[1].val
        val = selectBitRange(self.val, firstBitNo, size)
        vld = selectBitRange(self.vldMask, firstBitNo, size)

        retT = self._dtype.__class__(size, signed=self._dtype.signed)
        return self.__class__(val, retT, vld, updateTime=updateTime)

    def _getitem__val(self, key):
        # using self.__class__ because in simulator this method is called
        # for SimBits and we do not want to work with Bits in sim
        if isinstance(key._dtype, Integer):
            return self._getitem__val_int(key)
        elif key._dtype == SLICE:
            return self._getitem__val_slice(key)
        else:
            raise TypeError(key)

    def __getitem__(self, key):
        """
        [] operator

        :attention: Table below is for litle endian bit order (MSB:LSB)
            which is default.
            This is **reversed** as it is in pure python
            where it is [0, len(self)].

        :attention: slice on slice f signal is automatically reduced
            to single slice

        +-----------------------------+----------------------------------------------------------------------------------+
        | a[up:low]                   | items low through up; a[16:8] selects upper byte from 16b vector a               |
        +-----------------------------+----------------------------------------------------------------------------------+
        | a[up:]                      | low is automatically substituted with 0; a[8:] will select lower 8 bits          |
        +-----------------------------+----------------------------------------------------------------------------------+
        | a[:end]                     | up is automatically substituted; a[:8] will select upper byte from 16b vector a  |
        +-----------------------------+----------------------------------------------------------------------------------+
        | a[:], a[-1], a[-2:], a[:-2] | raises NotImplementedError   (not implemented due to complicated support in hdl) |
        +-----------+----------------------------------------------------------------------------------------------------+
        """
        iamVal = isinstance(self, Value)
        st = self._dtype
        length = st.bit_length()
        if length == 1 and not st.forceVector:
            # assert not indexing on single bit
            raise TypeError("indexing on single bit")

        if isinstance(key, slice):
            key = slice_to_SLICE(key, length)
            isSLICE = True
        else:
            isSLICE = isinstance(key, Slice.getValueCls())

        if isSLICE:
            # :note: downto notation
            start = key.val[0]
            stop = key.val[1]
            startIsVal = isinstance(start, Value)
            stopIsVal = isinstance(stop, Value)
            indexesAreValues = startIsVal and stopIsVal
        else:
            key = toHVal(key)

        iAmResultOfIndexing = (not iamVal and
                               hasattr(self, "origin") and
                               len(self.drivers) == 1 and
                               isinstance(self.origin, Operator) and
                               self.origin.operator == AllOps.INDEX)
        if isSLICE:
            if indexesAreValues and start.val == length and stop.val == 0:
                # selecting all bits no conversion needed
                return self

            if iAmResultOfIndexing:
                # try reduce self and parent slice to one
                original, parentIndex = self.origin.operands
                if isinstance(parentIndex._dtype, Slice):
                    parentLower = parentIndex.val[1]
                    start = start + parentLower
                    stop = stop + parentLower
                    return original[start:stop]

            # check start boundaries
            if startIsVal:
                _start = int(start)
                if _start < 0 or _start > length:
                    raise IndexError(_start, length)

            # check end boundaries
            if stopIsVal:
                _stop = int(stop)
                if _stop < 0 or _stop > length:
                    raise IndexError(_stop, length)

            # check width of selected range
            if startIsVal and stopIsVal and _start - _stop <= 0:
                raise IndexError(_start, _stop)

            if iamVal:
                return self._getitem__val(key)
            else:
                key = start._downto(stop)
                _resWidth = start - stop
                resT = Bits(width=_resWidth, forceVector=True,
                            signed=st.signed)

        elif isinstance(key, IntegerVal):
            # check index range
            _v = int(key)
            if _v < 0 or _v > length - 1:
                raise IndexError(_v)

            resT = BIT
            if iamVal:
                return self._getitem__val(key)

        elif isinstance(key, RtlSignalBase):
            t = key._dtype
            if isinstance(t, Integer):
                resT = BIT
            elif isinstance(t, Slice):
                resT = Bits(width=key.staticEval()._size(),
                            forceVector=st.forceVector, signed=st.signed)
            elif isinstance(t, Bits):
                resT = BIT
                key = key._auto_cast(INT)
            else:
                raise TypeError(
                    "Index operation not implemented"
                    " for index of type %r" % (t))

        else:
            raise TypeError(
                "Index operation not implemented for index %r" % (key))

        return Operator.withRes(AllOps.INDEX, [self, key], resT)

    def _setitem__val(self, index, value):
        if index._isFullVld():
            if index._dtype == SLICE:
                size = index._size()
                noOfFirstBit = index.val[1].val
                self.val = setBitRange(self.val, noOfFirstBit, size, value.val)
                self.vldMask = setBitRange(
                    self.vldMask, noOfFirstBit, size, value.vldMask)
            elif isinstance(index._dtype, Integer):
                self.val = bitSetTo(self.val, index.val, value.val)
                self.vldMask = bitSetTo(self.vldMask, index.val, value.vldMask)
            else:
                raise TypeError("Not implemented for index %r" % (index))
            self.updateTime = max(index.updateTime, value.updateTime)
        else:
            self.vldMask = 0

    def __setitem__(self, index, value):
        """this can not be called in desing description on non static values,
        only simulator can resolve this (in design use self[index] ** value
        instead of self[index] = value)
        """
        # convert index to hSlice or hInt
        indexConst = True
        if not isinstance(index, Value):
            if isinstance(index, RtlSignalBase):
                if index._const:
                    index = index.staticEval()
                else:
                    indexConst = False

            elif isinstance(index, slice):
                length = self._dtype.bit_length()
                index = slice_to_SLICE(index, length)
            else:
                index = hInt(index)
        if indexConst and not index._isFullVld():
            indexConst = False

        # convert value to bits of length specified by index
        if indexConst:
            if index._dtype == SLICE:
                itemT = Bits(index._size())
            else:
                itemT = BIT

            if not isinstance(value, Value):
                if isinstance(value, RtlSignalBase):
                    if value._const:
                        value = value.staticEval()._auto_cast(itemT)
                        valueConst = True
                    else:
                        valueConst = False
                else:
                    value = itemT.fromPy(value)
                    valueConst = True
            else:
                valueConst = True
                value = value._auto_cast(itemT)

        if indexConst and valueConst and isinstance(self, Value):
            return self._setitem__val(index, value)

        raise TypeError(
            "Only simulator can resolve []= for signals or invalid index")

    def _invert__val(self):
        v = self.clone()
        v.val = ~v.val
        w = v._dtype.bit_length()
        v.val &= mask(w)
        if self._dtype.signed:
            v.val = signFix(v.val, w)
        return v

    def __invert__(self):
        if isinstance(self, Value):
            return self._invert__val()
        else:
            try:
                # double negation
                d = self.singleDriver()
                if isinstance(d, Operator) and d.operator == AllOps.NOT:
                    return d.operands[0]
            except MultipleDriversExc:
                pass
            return Operator.withRes(AllOps.NOT, [self], self._dtype)

    # comparisons
    def _eq__val(self, other):
        return bitsCmp__val(self, other, AllOps.EQ, eq)

    def _eq(self, other):
        return bitsCmp(self, other, AllOps.EQ, eq)

    def _ne__val(self, other):
        return bitsCmp__val(self, other, AllOps.NEQ, ne)

    def __ne__(self, other):
        return bitsCmp(self, other, AllOps.NEQ)

    def _lt__val(self, other):
        return bitsCmp__val(self, other, AllOps.LT, lt)

    def __lt__(self, other):
        return bitsCmp(self, other, AllOps.LT)

    def _gt__val(self, other):
        return bitsCmp__val(self, other, AllOps.GT, gt)

    def __gt__(self, other):
        return bitsCmp(self, other, AllOps.GT)

    def _ge__val(self, other):
        return bitsCmp__val(self, other, AllOps.GE, ge)

    def __ge__(self, other):
        return bitsCmp(self, other, AllOps.GE)

    def _le__val(self, other):
        return bitsCmp__val(self, other, AllOps.LE, le)

    def __le__(self, other):
        return bitsCmp(self, other, AllOps.LE)

    def _xor__val(self, other):
        return bitsBitOp__val(self, other, AllOps.XOR, vldMaskForXor)

    def __xor__(self, other):
        return bitsBitOp(self, other, AllOps.XOR, vldMaskForXor, tryReduceXor)

    def _and__val(self, other):
        return bitsBitOp__val(self, other, AllOps.AND, vldMaskForAnd)

    def __and__(self, other):
        return bitsBitOp(self, other, AllOps.AND, vldMaskForAnd, tryReduceAnd)

    def _or__val(self, other):
        return bitsBitOp__val(self, other, AllOps.OR, vldMaskForOr)

    def __or__(self, other):
        return bitsBitOp(self, other, AllOps.OR, vldMaskForOr, tryReduceOr)

    def _sub__val(self, other):
        return bitsArithOp__val(self, other, AllOps.SUB)

    def __sub__(self, other):
        return bitsArithOp(self, other, AllOps.SUB)

    def _add__val(self, other):
        return bitsArithOp__val(self, other, AllOps.ADD)

    def __add__(self, other):
        return bitsArithOp(self, other, AllOps.ADD)

    def _mul__val(self, other):
        # [TODO] resT should be wider
        resT = self._dtype
        v = self.val * other.val
        v &= resT.all_mask()
        if resT.signed:
            v = signFix(v, resT.bit_length())

        result = resT.fromPy(v)
        if not self._isFullVld() or not other._isFullVld():
            result.vldMask = 0
        result.updateTime = max(self.updateTime, other.updateTime)
        return result

    def __mul__(self, other):
        other = toHVal(other)
        if not isinstance(other._dtype, (Integer, Bits)):
            raise TypeError(other)

        if areValues(self, other):
            return self._mul__val(other)
        else:
            myT = self._dtype
            if self._dtype.signed is None:
                self = self._unsigned()

            if isinstance(other._dtype, Bits):
                s = other._dtype.signed
                if s is None:
                    other = other._unsigned()

            elif isinstance(other._dtype, Integer):
                pass
            else:
                raise TypeError("%r %r %r" % (self, AllOps.MUL, other))

            if isinstance(other._dtype, Integer):
                res_w = myT.bit_length() * 2
                res_sign = self._dtype.signed
            else:
                res_w = myT.bit_length() + other._dtype.bit_length()
                res_sign = self._dtype.signed or other._dtype.signed

            subResT = Bits(res_w, signed=res_sign)
            o = Operator.withRes(AllOps.MUL, [self, other], subResT)
            resT = Bits(res_w, signed=myT.signed)
            return o._auto_cast(resT)



"""
This package contains classes for representation of HDL
languages (hdl types, RtlSignal, Assignment, HWProcess etc.).

It also contains classes for represenation of complex HW strucures
like dense transaction template (TransTmpl, FrameTmpl).
"""



from jinja2.environment import Environment
from jinja2.loaders import PackageLoader

from hwt.hdl.constants import DIRECTION
from hwt.hdl.entity import Entity
from hwt.interfaces.std import Clk
from hwt.serializer.generic.serializer import GenericSerializer
from hwt.serializer.generic.nameScope import LangueKeyword
from hwt.serializer.systemC.context import SystemCCtx
from hwt.serializer.systemC.keywords import SYSTEMC_KEYWORDS
from hwt.serializer.systemC.ops import SystemCSerializer_ops
from hwt.serializer.systemC.statements import SystemCSerializer_statements
from hwt.serializer.systemC.type import SystemCSerializer_type
from hwt.serializer.systemC.value import SystemCSerializer_value
from hwt.serializer.utils import maxStmId


class SystemCSerializer(SystemCSerializer_value, SystemCSerializer_type,
                        SystemCSerializer_statements, SystemCSerializer_ops,
                        GenericSerializer):
    """
    Serialized used to convert HWT design to SystemC code
    """
    fileExtension = '.cpp'
    _keywords_dict = {kw: LangueKeyword() for kw in SYSTEMC_KEYWORDS}
    env = Environment(loader=PackageLoader('hwt',
                                           'serializer/systemC/templates'))
    moduleTmpl = env.get_template('module.cpp.template')
    methodTmpl = env.get_template("method.cpp.template")
    ifTmpl = env.get_template("if.cpp.template")
    switchTmpl = env.get_template("switch.cpp.template")

    @classmethod
    def getBaseContext(cls):
        return SystemCCtx(cls.getBaseNameScope(), 0, None, None)

    @classmethod
    def comment(cls, comentStr):
        return "\n".join(["/*", comentStr, "*/"])

    @classmethod
    def PortItem(cls, p, ctx):
        d = cls.DIRECTION(p.direction)
        p.name = ctx.scope.checkedName(p.name, p)
        p.getSigInside().name = p.name
        if isinstance(p.getSigInside()._interface, Clk):
            return "sc_%s_clk %s;" % (d, p.name)

        return "sc_%s<%s> %s;" % (d,
                                  cls.HdlType(p._dtype, ctx),
                                  p.name)

    @classmethod
    def DIRECTION(cls, d):
        return d.name.lower()

    @classmethod
    def Entity(cls, ent, ctx):
        cls.Entity_prepare(ent, ctx)
        # [TODO] separate declarations from definitions
        doc = ent.__doc__
        if doc and id(doc) != id(Entity.__doc__):
            return cls.comment(doc) + "\n"
        return ""

    @classmethod
    def GenericItem(cls, g, ctx):
        # [TODO] params currently serialized evaluated
        return ""

    @classmethod
    def Architecture_var(cls, v, serializerVars, extraTypes,
                         extraTypes_serialized, ctx, childCtx):
        """
        :return: list of extra discovered processes
        """
        v.name = ctx.scope.checkedName(v.name, v)
        serializedVar = cls.SignalItem(v, childCtx, declaration=True)
        serializerVars.append(serializedVar)

    @classmethod
    def Architecture(cls, arch, ctx):
        serializerVars = []
        procs = []
        extraTypes = set()
        extraTypes_serialized = []
        arch.variables.sort(key=lambda x: (x.name, x._instId))
        arch.componentInstances.sort(key=lambda x: x._name)

        childCtx = ctx.withIndent()
        ports = list(map(lambda pi: cls.PortItem(
            pi, childCtx), arch.entity.ports))

        extraProcesses = []
        for v in arch.variables:
            cls.Architecture_var(v,
                                 serializerVars,
                                 extraTypes,
                                 extraTypes_serialized,
                                 ctx,
                                 childCtx)

        arch.processes.extend(extraProcesses)
        arch.processes.sort(key=lambda x: (x.name, maxStmId(x)))
        for p in arch.processes:
            procs.append(cls.HWProcess(p, childCtx))

        # architecture names can be same for different entities
        # arch.name = scope.checkedName(arch.name, arch, isGlobal=True)
        processesSensitivity = []
        sensitivityCtx = ctx.forSensitivityList()
        for p in arch.processes:
            sens = list(map(lambda s: cls.asHdl(
                s, sensitivityCtx), p.sensitivityList))
            processesSensitivity.append((p.name, sens))

        return cls.moduleTmpl.render(
            processesSensitivity=processesSensitivity,
            name=arch.getEntityName(),
            ports=ports,
            signals=serializerVars,
            extraTypes=extraTypes_serialized,
            processes=procs,
            processObjects=arch.processes,
            componentInstances=arch.componentInstances,
            DIRECTION=DIRECTION,
        )



from _random import Random
from collections import deque
from inspect import isgenerator
import os
import unittest

from hwt.hdl.constants import Time
from hwt.hdl.types.arrayVal import HArrayVal
from hwt.hdl.value import Value
from hwt.simulator.agentConnector import valToInt
from hwt.simulator.configVhdlTestbench import HdlSimConfigVhdlTestbench
from hwt.simulator.hdlSimulator import HdlSimulator
from hwt.simulator.shortcuts import simPrepare
from hwt.simulator.simSignal import SimSignal
from hwt.simulator.vcdHdlSimConfig import VcdHdlSimConfig


def allValuesToInts(sequenceOrVal):
    if isinstance(sequenceOrVal, HArrayVal):
        sequenceOrVal = sequenceOrVal.val

    if isinstance(sequenceOrVal, Value):
        return valToInt(sequenceOrVal)
    elif not sequenceOrVal:
        return sequenceOrVal
    elif (isinstance(sequenceOrVal, (list, tuple, deque))
          or isgenerator(sequenceOrVal)):
        seq = []
        for i in sequenceOrVal:
            seq.append(allValuesToInts(i))

        if isinstance(sequenceOrVal, tuple):
            return tuple(seq)

        return seq
    else:
        return sequenceOrVal


class SimTestCase(unittest.TestCase):
    """
    This is TestCase class contains methods which are usually used during
    hdl simulation.

    :attention: self.model, self.procs has to be specified before running
        doSim (you can use prepareUnit method)
    """
    _defaultSeed = 317

    def getTestName(self):
        className, testName = self.id().split(".")[-2:]
        return "%s_%s" % (className, testName)

    def doSim(self, time, name=None, config=None):
        if name is None:
            outputFileName = "tmp/" + self.getTestName() + ".vcd"
        else:
            outputFileName = name

        d = os.path.dirname(outputFileName)
        if d:
            os.makedirs(d, exist_ok=True)
        with open(outputFileName, 'w') as outputFile:
            sim = HdlSimulator()

            if config is None:
                # configure simulator to log in vcd
                config = VcdHdlSimConfig(outputFile)
            sim.config = config

            # run simulation, stimul processes are register after initial
            # initialization
            sim.simUnit(self.model, time=time, extraProcesses=self.procs)
            return sim

    def __serializeTestbenchDump(self, time, file):
        sim = HdlSimulator()

        # configure simulator to log in vcd
        sim.config = HdlSimConfigVhdlTestbench(self.u)

        # run simulation, stimul processes are register after initial
        # initialization
        sim.simUnit(self.model, time=time, extraProcesses=self.procs)

        sim.config.dump(file)

        return sim

    def dumpHdlTestbench(self, time, file=None):
        if file is not None:
            outputFileName = file
        else:
            outputFileName = "tmp/" + self.getTestName() + "_tb.vhd"

        if isinstance(file, str):
            d = os.path.dirname(outputFileName)
            if d:
                os.makedirs(d, exist_ok=True)
            with open(outputFileName, 'w') as outputFile:
                return self.__serializeTestbenchDump(time, outputFile)

        else:
            return self.__serializeTestbenchDump(time, file)

    def assertValEqual(self, first, second, msg=None):
        if isinstance(first, SimSignal):
            first = first._val
        if not isinstance(first, int) and first is not None:
            first = valToInt(first)

        return unittest.TestCase.assertEqual(self, first, second, msg=msg)

    def assertEmpty(self, val, msg=None):
        return unittest.TestCase.assertEqual(self, len(val), 0, msg=msg)

    def assertValSequenceEqual(self, seq1, seq2, msg=None, seq_type=None):
        """
        An equality assertion for ordered sequences (like lists and tuples).
        For the purposes of this function, a valid ordered sequence type is one
        which can be indexed, has a length, and has an equality operator.

        Args:

        :param seq1: can contain instance of values or nested list of them
        :param seq2: items are not converted
        :param seq_type: The expected datatype of the sequences, or None if no
            datatype should be enforced.
        :param msg: Optional message to use on failure instead of a list of
            differences.
        """
        seq1 = allValuesToInts(seq1)
        self.assertSequenceEqual(seq1, seq2, msg, seq_type)

    def simpleRandomizationProcess(self, agent):
        seed = self._rand.getrandbits(64)
        random = Random(seed)
        timeQuantum = 10 * Time.ns  # default clk period

        def randomEnProc(sim):
            # small space at start to modify agents when they are inactive
            yield sim.wait(timeQuantum / 4)
            while True:
                en = random.random() < 0.5
                if agent.getEnable() != en:
                    agent.setEnable(en, sim)
                delay = int(random.random() * 2) * timeQuantum
                yield sim.wait(delay)
        return randomEnProc

    def randomize(self, intf):
        """
        Randomly disable and enable interface for testing purposes
        """
        randomEnProc = self.simpleRandomizationProcess(intf._ag)
        self.procs.append(randomEnProc)

    def prepareUnit(self, unit, modelCls=None, dumpModelIn=None,
                    onAfterToRtl=None, targetPlatform=None):
        """
        Create simulation model and connect it with interfaces of original unit
        and decorate it with agents and collect all simulation processes

        :param unit: interface level unit which you wont prepare for simulation
        :param modelCls: class of rtl simulation model to run simulation on,
            if is None rtl sim model will be generated from unit
        :param dumpModelIn: folder to where put sim model files (if is None
            sim model will be constructed only in memory)
        :param onAfterToRtl: callback fn(unit) which will be called unit after
            it will be synthesised to rtl
        """
        self.u, self.model, self.procs = simPrepare(
            unit,
            modelCls=modelCls,
            targetPlatform=targetPlatform,
            dumpModelIn=dumpModelIn,
            onAfterToRtl=onAfterToRtl)

    def setUp(self):
        self._rand = Random(self._defaultSeed)



from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import INT, BOOL, STR, BIT


def hInt(pyVal):
    """ create hdl integer value (for example integer value in vhdl)"""
    return INT.fromPy(pyVal)


def hBool(pyVal):
    """ create hdl bool value (for example bool value in vhdl)"""
    return BOOL.fromPy(pyVal)


def hStr(pyVal):
    """create hdl string value (for example string value in vhdl)"""
    return STR.fromPy(pyVal)


def hBit(pyVal):
    """create hdl bit value (for example STD_LOGIC value in vhdl)"""
    return BIT.fromPy(pyVal)


def vec(val, width, signed=None):
    """create hdl vector value"""
    return Bits(width, signed, forceVector=True).fromPy(val)



VHLD_KEYWORDS = [
    "abs", "access", "across", "after", "alias", "all", "and", "architecture",
    "array", "assert", "attribute", "begin", "block", "body", "break",
    "bugger", "bus", "case", "component", "configuration", "constant",
    "disconnect", "downto", "end", "entity", "else", "elsif", "exit", "file",
    "for", "function", "generate", "generic", "group", "guarded", "if",
    "impure", "in", "inertial", "inout", "is", "label", "library", "limit",
    "linkage", "literal", "loop", "map", "mod", "nand", "nature", "new",
    "next", "noise", "nor", "not", "null", "of", "on", "open", "or", "others",
    "out", "package", "port", "postponed", "process", "procedure",
    "procedural", "pure", "quantity", "range", "reverse_range", "reject",
    "rem", "record", "reference", "register", "report", "return", "rol", "ror",
    "select", "severity", "shared", "signal", "sla", "sll", "spectrum", "sra",
    "srl", "subnature", "subtype", "terminal", "then", "through", "to",
    "tolerance", "transport", "type", "unaffected", "units", "until", "use",
    "variable", "wait", "with", "when", "while", "xnor", "xor"]



from hwt.bitmask import mask
from hwt.hdl.types.hdlType import HdlType
from hwt.serializer.generic.indent import getIndent


class Bits(HdlType):
    """
    Elemental HDL type representing bits (vector or single bit)
    """
    def __init__(self, width, signed=None, forceVector=False, negated=False):
        """
        :param negated: if true the value is in negated form
        :param forceVector: use always hdl vector type
            (for example std_logic_vector(0 downto 0)
             instead of std_logic in VHDL)
        """
        self.forceVector = forceVector
        self.negated = negated

        w = int(width)
        assert isinstance(w, int) and w > 0
        self._widthVal = w

        self.signed = signed
        self.width = width

        self._allMask = mask(self._widthVal)

    def __eq__(self, other):
        if self is other:
            return True

        if self._widthVal == 1:
            return (isinstance(other, Bits)
                    and other._widthVal == 1
                    and self.signed == other.signed
                    and self.forceVector == other.forceVector)
        else:
            return (isinstance(other, Bits)
                    and other._widthVal == self._widthVal
                    and self.signed == other.signed)

    def __hash__(self):
        return hash((self.signed, self._widthVal, self.forceVector))

    def all_mask(self):
        """
        :return: mask for bites of this type ( 0b111 for Bits(3) )
        """
        return self._allMask

    def bit_length(self):
        """
        :return: bit width for this type
        """
        return self._widthVal

    @classmethod
    def get_auto_cast_fn(cls):
        from hwt.hdl.types.bitsCast import convertBits
        return convertBits

    @classmethod
    def get_reinterpret_cast_fn(cls):
        from hwt.hdl.types.bitsCast import reinterpretBits
        return reinterpretBits

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.bitsVal import BitsVal
            cls._valCls = BitsVal
            return cls._valCls

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not None is used as a additional
            information about on which address this type is stored
            (used only by HStruct)
        :param expandStructs: expand HStructTypes (used by HStruct and HArray)
        """
        c = self.width
        if isinstance(c, int):
            constr = "%dbits" % c
        else:
            from hwt.serializer.vhdl.serializer import VhdlSerializer
            ctx = VhdlSerializer.getBaseContext()
            constr = VhdlSerializer.asHdl(self.width, ctx)
            constr = "%s, %dbits" % (constr, self.bit_length())

        if self.signed:
            constr += ", signed"
        elif self.signed is False:
            constr += ", unsigned"

        return "%s<%s, %s>" % (getIndent(indent),
                               self.__class__.__name__,
                               constr)



from typing import Union

from hwt.simulator.agentBase import AgentBase
from hwt.synthesizer.interface import Interface
from hwt.synthesizer.interfaceLevel.interfaceUtils.proxy import InterfaceProxy


def getMonitors(intf: Union[Interface, InterfaceProxy]):
    if intf._arrayElemCache or (isinstance(intf, InterfaceProxy)
                                and intf._origIntf._arrayElemCache):
        for p in intf:
            yield from p._ag.getMonitors()
    else:
        yield from intf._ag.getMonitors()


def getDrivers(intf: Union[Interface, InterfaceProxy]):
    if intf._arrayElemCache or (isinstance(intf, InterfaceProxy)
                                and intf._origIntf._arrayElemCache):
        for p in intf:
            yield from p._ag.getDrivers()
    else:
        yield from intf._ag.getDrivers()


class StructIntfAgent(AgentBase):
    """
    Agent for StructIntf inteface

    :summary: only purpose is to instantiate agents for child interfaces
    """

    def __init__(self, intf):
        AgentBase.__init__(self, intf)
        # if interface is InterfaceProxy and this proxy is container
        # of another array we generate agents on items in array
        # otherwise generate agents for each subinterface

        for intf in intf._interfaces:
            if intf._isInterfaceArray():
                for p in intf:
                    p._initSimAgent()
            else:
                intf._initSimAgent()

    def getMonitors(self):
        for intf in self.intf._interfaces:
            yield from getMonitors(intf)

    def getDrivers(self):
        for intf in self.intf._interfaces:
            yield from getDrivers(intf)



from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.bool import HBool
from hwt.hdl.types.defs import INT
from hwt.hdl.types.hdlType import default_auto_cast_fn
from hwt.hdl.types.struct import HStruct
from hwt.hdl.types.union import HUnion
from hwt.hdl.value import Value
from hwt.synthesizer.vectorUtils import iterBits, fitTo_t


def convertBits__val(self, val, toType):
    if isinstance(toType, HBool):
        return val._eq(self.getValueCls().fromPy(1, self))
    elif isinstance(toType, Bits):
        return val._convSign__val(toType.signed)
    elif toType == INT:
        return INT.getValueCls()(val.val,
                                 INT,
                                 int(val._isFullVld()),
                                 val.updateTime)

    return default_auto_cast_fn(self, val, toType)


def convertBits(self, sigOrVal, toType):
    """
    Cast signed-unsigned, to int or bool
    """
    if isinstance(sigOrVal, Value):
        return convertBits__val(self, sigOrVal, toType)
    elif isinstance(toType, HBool):
        if self.bit_length() == 1:
            v = 0 if sigOrVal._dtype.negated else 1
            return sigOrVal._eq(self.getValueCls().fromPy(v, self))
    elif isinstance(toType, Bits):
        if self.bit_length() == toType.bit_length():
            return sigOrVal._convSign(toType.signed)
    elif toType == INT:
        return Operator.withRes(AllOps.BitsToInt, [sigOrVal], toType)

    return default_auto_cast_fn(self, sigOrVal, toType)


def reinterpret_bits_to_hstruct(sigOrVal, hStructT):
    """
    Reinterpret signal of type Bits to signal of type HStruct
    """
    container = hStructT.fromPy(None)
    offset = 0
    for f in hStructT.fields:
        t = f.dtype
        width = t.bit_length()
        if f.name is not None:
            s = sigOrVal[(width + offset):offset]
            s = s._reinterpret_cast(t)
            setattr(container, f.name, s)

        offset += width

    return container


def reinterpret_bits_to_harray(sigOrVal, hArrayT):
    elmT = hArrayT.elmType
    elmWidth = elmT.bit_length()
    a = hArrayT.fromPy(None)
    for i, item in enumerate(iterBits(sigOrVal,
                                      bitsInOne=elmWidth,
                                      skipPadding=False)):
        item = item._reinterpret_cast(elmT)
        a[i] = item

    return a


def reinterpretBits__val(self, val, toType):
    if isinstance(toType, HStruct):
        return reinterpret_bits_to_hstruct(val, toType)
    elif isinstance(toType, HUnion):
        raise NotImplementedError()
    elif isinstance(toType, HArray):
        return reinterpret_bits_to_harray(val, toType)

    return default_auto_cast_fn(self, val, toType)


def reinterpretBits(self, sigOrVal, toType):
    """
    Cast object of same bit size between to other type
    (f.e. bits to struct, union or array)
    """
    if isinstance(sigOrVal, Value):
        return reinterpretBits__val(self, sigOrVal, toType)
    elif isinstance(toType, Bits):
        return fitTo_t(sigOrVal, toType)
    elif sigOrVal._dtype.bit_length() == toType.bit_length():
        if isinstance(toType, HStruct):
            raise reinterpret_bits_to_hstruct(sigOrVal, toType)
        elif isinstance(toType, HUnion):
            raise NotImplementedError()
        elif isinstance(toType, HArray):
            reinterpret_bits_to_harray(sigOrVal, toType)

    return default_auto_cast_fn(self, sigOrVal, toType)



from hwt.hdl.hdlObject import HdlObject
from hwt.hdl.value import Value
from hwt.pyUtils.arrayQuery import arr_all
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal, RtlSignalBase


def getCtxFromOps(operands):
    for o in operands:
        if isinstance(o, RtlSignalBase):
            return o.ctx
    raise TypeError("Can not find context because there is no signal in ops"
                    "(value operators should be already resolved)")


def isConst(item):
    return isinstance(item, Value) or item._const


class Operator(HdlObject):
    """
    Class of operator in expression tree

    :ivar operands: list of operands
    :ivar evalFn: function to evaluate this operator
    :ivar operator: OpDefinition instance
    :ivar result: result signal of this operator
    """

    def __init__(self, operator, operands):
        self.operands = list(operands)
        self.operator = operator
        self.result = None

    def registerSignals(self, outputs=[]):
        """
        Register potential signals to drivers/endpoints
        """
        for o in self.operands:
            if isinstance(o, RtlSignalBase):
                if o in outputs:
                    o.drivers.append(self)
                else:
                    o.endpoints.append(self)
            elif isinstance(o, Value):
                pass
            else:
                raise NotImplementedError(
                    "Operator operands can be"
                    " only signal or values got:%r" % (o))

    def staticEval(self):
        """
        Recursively statistically evaluate result of this operator
        """
        for o in self.operands:
            o.staticEval()
        self.result._val = self.evalFn()

    def evalFn(self, simulator=None):
        """
        Syntax sugar
        """
        return self.operator.eval(self, simulator=simulator)

    def __eq__(self, other):
        return self is other or (
            type(self) is type(other) and
            self.operator == other.operator and
            self.operands == other.operands
        )

    @staticmethod
    def withRes(opDef, operands, resT, outputs=[]):
        """
        Create operator with result signal

        :ivar resT: data type of result signal
        :ivar outputs: iterable of singnals which are outputs
            from this operator
        """
        op = Operator(opDef, operands)
        out = RtlSignal(getCtxFromOps(operands), None, resT)
        out._const = arr_all(op.operands, isConst)
        out.drivers.append(op)
        out.origin = op
        op.result = out
        op.registerSignals(outputs)
        if out._const:
            out.staticEval()
        return out

    def __hash__(self):
        return hash((self.operator, frozenset(self.operands)))



from hwt.code import Concat
from hwt.synthesizer.vectorUtils import iterBits


def reverseByteOrder(signalOrVal):
    """
    Reverse byteorder (littleendian/bigendian) of signal or value
    """
    w = signalOrVal._dtype.bit_length()
    i = w
    items = []

    while i > 0:
        # take last 8 bytes or rest
        lower = max(i - 8, 0)
        items.append(signalOrVal[i:lower])
        i -= 8

    return Concat(*items)


def reversedBits(sigOrVal):
    return Concat(* list(iterBits(sigOrVal)))



from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps


def getBaseCond(c):
    """
    if is negated return original cond and negated flag
    """
    isNegated = False
    try:
        drivers = c.drivers
    except AttributeError:
        return (c, isNegated)

    if len(drivers) == 1:
        d = list(c.drivers)[0]
        if isinstance(d, Operator) and d.operator == AllOps.NOT:
            c = d.operands[0]
            isNegated = True

    return (c, isNegated)



from copy import copy

from hwt.serializer.generic.context import SerializerCtx
from hwt.serializer.verilog.utils import verilogTypeOfSig, SIGNAL_TYPE


class VerilogSerializerCtx(SerializerCtx):
    """
    :ivar signalType: member of SIGNAL_TYPE
    """

    def forPort(self):
        ctx = copy(self)
        ctx.signalType = SIGNAL_TYPE.PORT
        return ctx

    def forSignal(self, signalItem):
        ctx = copy(self)
        ctx.signalType = verilogTypeOfSig(signalItem)
        return ctx





class UniqList(list):
    """
    List of unique items
    """

    def __init__(self, *args, **kwargs):
        super(UniqList, self).__init__(*args, **kwargs)
        self.__s = set()

    def append(self, item):
        if item in self.__s:
            return False
        else:
            self.__s.add(item)
            list.append(self, item)
            return True
    
    def extend(self, items):
        for item in items:
            self.append(item)

    def remove(self, item):
        self.__s.remove(item)
        return list.remove(self, item)

    def __contains__(self, key):
        return key in self.__s



from hwt.hdl.constants import INTF_DIRECTION
from hwt.synthesizer.unit import Unit
from hwt.synthesizer.exceptions import IntfLvlConfErr


def setOut(*interfaces):
    """
    set interfaces as output interfaces

    :attention: only for instances of EmptyUnit
    """
    for i in interfaces:
        i._setDirectionsLikeIn(INTF_DIRECTION.SLAVE)


class EmptyUnit(Unit):
    """
    Unit used for prototyping all output interfaces are connected
    to _defaultValue and this is only think which architecture contains

    :cvar _defaultValue: this value is used to initialize all signals
    """
    _defaultValue = None

    def _toRtl(self, targetPlatform):
        assert not self._wasSynthetised()
        self._targetPlatform = targetPlatform

        if not hasattr(self, "_name"):
            self._name = self._getDefaultName()
        for i in self._interfaces:
            i._setDirectionsLikeIn(INTF_DIRECTION.MASTER)

        self._loadMyImplementations()
        # construct params for entity (generics)
        self._cntx.params = self._buildParams()
        externInterf = []
        # prepare connections
        for i in self._interfaces:
            signals = i._signalsForInterface(self._cntx)
            if not i._isExtern:
                raise IntfLvlConfErr(
                    "All interfaces in EmptyUnit has to be extern, %s: %s is not"
                    % (self.__class__.__name__, i._getFullName()))
            externInterf.extend(signals)
            # i._resolveDirections()
            # connect outputs to dummy value
            for s in signals:
                if s._interface._direction == INTF_DIRECTION.SLAVE:
                    s(s._dtype.fromPy(self._defaultValue))

        if not externInterf:
            raise IntfLvlConfErr(
                "Can not find any external interface for unit %s"
                "- unit without interfaces are not allowed"
                % self._name)
        yield from self._synthetiseContext(externInterf)
        # self._checkEntityPortDirections()



from hwt.serializer.ip_packager.interfaces.intfConfig import IntfConfig, DEFAULT_CLOCK
from hwt.interfaces.std import Rst, Rst_n
from hwt.pyUtils.arrayQuery import where


class IP_Clk(IntfConfig):
    def __init__(self):
        super().__init__()
        self.name = "clock"
        self.version = "1.0"
        self.vendor = "xilinx.com"
        self.library = "signal"
        self.map = 'CLK'

    def postProcess(self, component, entity, allInterfaces, thisIf):
        rst = where(allInterfaces, lambda intf: (isinstance(intf, Rst_n) or
                                                 isinstance(intf, Rst)))
        rst = list(rst)
        if len(rst) > 0:
            rst = rst[0]
            self.addSimpleParam(thisIf, "ASSOCIATED_RESET",
                                rst._name)  # getResetPortName

        elif len(rst) > 1:
            raise Exception("Don't know how to work with multiple resets")

        intfs = where(
            allInterfaces, lambda intf: intf is not rst and intf is not self)
        self.addSimpleParam(thisIf, "ASSOCIATED_BUSIF", ":".join(
            map(lambda intf: intf._name, intfs)))
        self.addSimpleParam(thisIf, "FREQ_HZ", str(DEFAULT_CLOCK))


class IP_Rst(IntfConfig):
    def __init__(self):
        super().__init__()
        self.name = "reset"
        self.version = "1.0"
        self.vendor = "xilinx.com"
        self.library = "signal"
        self.map = "rst"

    def postProcess(self, component, entity, allInterfaces, thisIf):
        self.addSimpleParam(thisIf, "POLARITY", "ACTIVE_HIGH")


class IP_Rst_n(IntfConfig):
    def __init__(self):
        super().__init__()
        self.name = "reset"
        self.version = "1.0"
        self.vendor = "xilinx.com"
        self.library = "signal"
        self.map = "rst"

    def postProcess(self, component, entity, allInterfaces, thisIf):
        self.addSimpleParam(thisIf, "POLARITY", "ACTIVE_LOW")


class IP_Handshake(IntfConfig):
    def __init__(self):
        super().__init__()
        self.name = "handshake"
        self.version = "1.0"
        self.vendor = "hwt"
        self.library = "user"
        self.map = {"vld": "valid",
                    "rd": "ready",
                    "data": "data"}


class IP_BlockRamPort(IntfConfig):
    def __init__(self):
        super().__init__()
        self.name = "bram"
        self.version = "1.0"
        self.vendor = "xilinx.com"
        self.library = "interface"
        self.map = {'addr': "ADDR",
                    "clk": 'CLK',
                    'din': "DIN",
                    'dout': "DOUT",
                    'en': "EN",
                    'we': "WE",
                    }



from hwt.hdl.constants import INTF_DIRECTION, DIRECTION
from hwt.synthesizer.exceptions import IntfLvlConfErr


class InterfaceDirectionFns():

    def _setDirectionsLikeIn(self, intfDir):
        # [TODO] array elements
        d = DIRECTION.asIntfDirection(self._masterDir)
        if intfDir == INTF_DIRECTION.MASTER or d == INTF_DIRECTION.TRISTATE:
            self._direction = d
            for i in self._interfaces:
                i._setDirectionsLikeIn(d)
        else:
            opDir = INTF_DIRECTION.opposite(d)
            self._direction = opDir
            for i in self._interfaces:
                i._setDirectionsLikeIn(opDir)

    def __directionProbe(self):
        if not self._interfaces:
            s = self._sig
            return (not bool(s.drivers), bool(s.drivers))

        allInMasterConf = True
        allInSlaveConf = True
        for i in self._interfaces:
            i._resolveDirections(updateDir=False)
            d = i._direction
            md = DIRECTION.asIntfDirection(i._masterDir)
            if d != INTF_DIRECTION.UNKNOWN:
                isLikeInM = d == md
                isLikeInS = d == INTF_DIRECTION.opposite(md)
                allInMasterConf = allInMasterConf and isLikeInM
                allInSlaveConf = allInSlaveConf and isLikeInS
        return (allInMasterConf, allInSlaveConf)

    def _resolveDirections(self, updateDir=True):
        allM, allS = self.__directionProbe()

        if allM and allS and self._arrayElemCache:  # if direction is nod clear from this intf. and it has elems.
            allM, allS = self._arrayElemCache[0].__directionProbe()

        if allM and allS:
            self._direction = INTF_DIRECTION.UNKNOWN
        elif allM:
            self._direction = INTF_DIRECTION.MASTER
        elif allS:
            self._direction = INTF_DIRECTION.SLAVE
        else:
            raise IntfLvlConfErr("Subinterfaces on %r \nhave not consistent directions\n%s" %
                                 (self,
                                  '\n'.join([str((i._direction, repr(i)))
                                             for i in self._interfaces])))

        if updateDir:
            if self._direction == INTF_DIRECTION.UNKNOWN:
                self._direction = INTF_DIRECTION.MASTER
            self._setDirectionsLikeIn(self._direction)

    def _setAsExtern(self, isExtern):
        """Set interface as extern"""
        self._isExtern = isExtern
        for prop in self._interfaces:
            prop._setAsExtern(isExtern)

    def _setDirLock(self, dirLock):
        self._dirLocked = dirLock
        for intf in self._interfaces:
            intf._setDirLock(dirLock)

    def _reverseDirection(self):
        """Reverse direction of this interface in implementation stage"""
        if self._dirLocked:
            raise IntfLvlConfErr("Can not reverse direction on interface %s because it was locked (%s)" %
                                 (repr(self), self._direction))
        self._direction = INTF_DIRECTION.opposite(self._direction)
        for intf in self._interfaces:
            intf._reverseDirection()



from hwt.hdl.entity import Entity
from hwt.hdl.types.array import HArray
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.generic.serializer import GenericSerializer
from hwt.serializer.generic.indent import getIndent
from hwt.serializer.generic.mapExpr import MapExpr
from hwt.serializer.generic.nameScope import LangueKeyword
from hwt.serializer.generic.portMap import PortMap
from hwt.serializer.utils import maxStmId
from hwt.serializer.verilog.context import VerilogSerializerCtx
from hwt.serializer.verilog.keywords import VERILOG_KEYWORDS
from hwt.serializer.verilog.ops import VerilogSerializer_ops
from hwt.serializer.verilog.statements import VerilogSerializer_statements
from hwt.serializer.verilog.tmplContainer import VerilogTmplContainer
from hwt.serializer.verilog.types import VerilogSerializer_types
from hwt.serializer.verilog.utils import SIGNAL_TYPE, verilogTypeOfSig
from hwt.serializer.verilog.value import VerilogSerializer_Value
from hwt.synthesizer.param import getParam
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.process import HWProcess
from hwt.hdl.statements import SwitchContainer
from hwt.hdl.types.typeCast import toHVal


class VerilogSerializer(VerilogTmplContainer, VerilogSerializer_types,
                        VerilogSerializer_Value, VerilogSerializer_statements,
                        VerilogSerializer_ops, GenericSerializer):
    _keywords_dict = {kw: LangueKeyword() for kw in VERILOG_KEYWORDS}
    fileExtension = '.v'

    @classmethod
    def getBaseContext(cls):
        return VerilogSerializerCtx(cls.getBaseNameScope(), 0, None, None)

    @classmethod
    def Entity(cls, ent, ctx):
        generics, ports = cls.Entity_prepare(ent, ctx)

        entVerilog = cls.moduleHeadTmpl.render(
            indent=getIndent(ctx.indent),
            name=ent.name,
            ports=ports,
            generics=generics
        )

        doc = ent.__doc__
        if doc and id(doc) != id(Entity.__doc__):
            doc = cls.comment(doc) + "\n"
            return doc + entVerilog
        else:
            return entVerilog

    @classmethod
    def hardcodeRomIntoProcess(cls, rom):
        """
        Due to verilog restrictions it is not posible to use array constants
        and rom memories has to be hardcoded as process
        """
        processes = []
        signals = []
        for e in rom.endpoints:
            assert isinstance(e, Operator) and e.operator == AllOps.INDEX, e
            me, index = e.operands
            assert me is rom

            # construct output of the rom
            romValSig = rom.ctx.sig(rom.name, dtype=e.result._dtype)
            signals.append(romValSig)
            romValSig.hidden = False

            # construct process which will represent content of the rom
            cases = [(toHVal(i), romValSig(v))
                     for i, v in enumerate(rom.defaultVal.val)]
            statements = [SwitchContainer(index, cases), ]

            p = HWProcess(rom.name, statements, {index, },
                          {index, }, {romValSig, })
            processes.append(p)

            # override usage of original index operator on rom
            # to use signal generated from this process
            def replaceOrigRomIndexExpr(x):
                if x is e.result:
                    return romValSig
                else:
                    return x
            for _e in e.result.endpoints:
                _e.operands = tuple(map(replaceOrigRomIndexExpr, _e.operands))
                e.result = romValSig

        return processes, signals

    @classmethod
    def Architecture_var(cls, v, serializerVars, extraTypes,
                         extraTypes_serialized, ctx, childCtx):
        """
        :return: list of extra discovered processes
        """
        t = v._dtype
        # if type requires extra definition
        if isinstance(t, HArray) and v.defaultVal.vldMask:
            if v.drivers:
                raise SerializerException("Verilog does not support RAMs"
                                          " with initialized value")
            eProcs, eVars = cls.hardcodeRomIntoProcess(v)
            for _v in eVars:
                _procs = cls.Architecture_var(_v, serializerVars, extraTypes,
                                              extraTypes_serialized, ctx,
                                              childCtx)
                eProcs.extend(_procs)
            return eProcs

        v.name = ctx.scope.checkedName(v.name, v)
        serializedVar = cls.SignalItem(v, childCtx, declaration=True)
        serializerVars.append(serializedVar)

        return []

    @classmethod
    def Architecture(cls, arch, ctx):
        serializerVars = []
        procs = []
        extraTypes = set()
        extraTypes_serialized = []
        arch.variables.sort(key=lambda x: (x.name, x._instId))
        arch.componentInstances.sort(key=lambda x: x._name)

        childCtx = ctx.withIndent()
        extraProcesses = []
        for v in arch.variables:
            _eProc = cls.Architecture_var(v,
                                          serializerVars,
                                          extraTypes,
                                          extraTypes_serialized,
                                          ctx,
                                          childCtx)

            extraProcesses.extend(_eProc)

        arch.processes.extend(extraProcesses)
        arch.processes.sort(key=lambda x: (x.name, maxStmId(x)))
        for p in arch.processes:
            procs.append(cls.HWProcess(p, childCtx))

        # architecture names can be same for different entities
        # arch.name = scope.checkedName(arch.name, arch, isGlobal=True)
        componentInstances = list(
            map(lambda c: cls.ComponentInstance(c, childCtx),
                arch.componentInstances))

        return cls.moduleBodyTmpl.render(
            indent=getIndent(ctx.indent),
            entityName=arch.getEntityName(),
            name=arch.name,
            variables=serializerVars,
            extraTypes=extraTypes_serialized,
            processes=procs,
            componentInstances=componentInstances
        )

    @classmethod
    def ComponentInstance(cls, entity, ctx):
        portMaps = []
        for pi in entity.ports:
            pm = PortMap.fromPortItem(pi)
            portMaps.append(pm)

        genericMaps = []
        for g in entity.generics:
            gm = MapExpr(g, g._val)
            genericMaps.append(gm)

        if len(portMaps) == 0:
            raise SerializerException("Incomplete component instance")

        # [TODO] check component instance name
        return cls.componentInstanceTmpl.render(
            indent=getIndent(ctx.indent),
            instanceName=entity._name,
            entity=entity,
            portMaps=[cls.PortConnection(x, ctx) for x in portMaps],
            genericMaps=[cls.MapExpr(x, ctx) for x in genericMaps]
        )

    @classmethod
    def comment(cls, comentStr):
        return "\n".join(["/*", comentStr, "*/"])

    @classmethod
    def GenericItem(cls, g, ctx):
        s = "%s %s" % (cls.HdlType(g._dtype, ctx.forPort()), g.name)
        if g.defaultVal is None:
            return s
        else:
            return ("parameter %s = %s"
                    % (s, cls.Value(getParam(g.defaultVal).staticEval(), ctx)))

    @classmethod
    def PortItem(cls, pi, ctx):
        t = cls.HdlType(pi._dtype, ctx.forPort())
        if verilogTypeOfSig(pi.getSigInside()) == SIGNAL_TYPE.REG:
            if t:
                f = "%s reg %s %s"
            else:
                f = "%s reg %s"
        else:
            if t:
                f = "%s %s %s"
            else:
                f = "%s %s"

        if t:
            return f % (cls.DIRECTION(pi.direction),
                        t, pi.name)
        else:
            return f % (cls.DIRECTION(pi.direction),
                        pi.name)

    @classmethod
    def PortConnection(cls, pc, ctx):
        if pc.portItem._dtype != pc.sig._dtype:
            raise SerializerException(
                "Port map %s is nod valid (types does not match)  (%s, %s)" % (
                    "%s => %s" % (pc.portItem.name, cls.asHdl(pc.sig, ctx)),
                    repr(pc.portItem._dtype), repr(pc.sig._dtype)))
        return ".%s(%s)" % (pc.portItem.name, cls.asHdl(pc.sig, ctx))

    @classmethod
    def MapExpr(cls, m, createTmpVar):
        return ".%s(%s)" % (m.compSig.name, cls.asHdl(m.value, createTmpVar))



from hwt.hdl.types.hdlType import HdlType
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class HArray(HdlType):
    """
    HDL array type

    :ivar elmType: type of elements
    :ivar size: number of items
    """

    def __init__(self, elmType, size):
        super(HArray, self).__init__()
        self.elmType = elmType
        self.size = size

    def __eq__(self, other):
        return self is other or (
            type(self) is type(other) and
            self.size == other.size and
            self.elmType == other.elmType
        )

    def __hash__(self):
        return hash(id(self))

    def bit_length(self):
        """
        :return: bit width for this type
        """
        try:
            itemSize = self.elmType.bit_length
        except AttributeError:
            itemSize = None
        if itemSize is None:
            raise TypeError(
                "Can not determine size of array because item has"
                " not determinable size")

        s = self.size
        if isinstance(s, RtlSignalBase):
            s = int(s.staticEval())
        return s * itemSize()

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.arrayVal import HArrayVal
            cls._valCls = HArrayVal
            return cls._valCls

    @classmethod
    def get_reinterpret_cast_fn(cls):
        from hwt.hdl.types.arrayCast import reinterpret_cast_harray
        return reinterpret_cast_harray

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not None is used as a additional
            information about on which address this type is stored
            (used only by HStruct)
        :param expandStructs: expand HStructTypes (used by HStruct and Array)
        """
        return "%s[%r]" % (self.elmType.__repr__(indent=indent,
                                                 withAddr=withAddr,
                                                 expandStructs=expandStructs),
                           self.size)





class HdlObject():
    """
    Base Hdl object class for object which can be directly serialized
    to target HDL language
    """

    def __repr__(self):
        from hwt.serializer.hwt.serializer import HwtSerializer
        ctx = HwtSerializer.getBaseContext()
        return getattr(HwtSerializer, self.__class__.__name__)(self, ctx)



from hwt.hdl.types.defs import INT
from hwt.hdl.types.hdlType import default_auto_cast_fn
from hwt.simulator.types.simInt import SIM_INT


def convertSimBits__val(self, sigOrVal, toType):
    if toType == INT or toType == SIM_INT:
        if self.signed:
            raise NotImplementedError()
        else:
            fullMask = self._allMask
            return INT.getValueCls()(sigOrVal.val,
                                     INT,
                                     sigOrVal.vldMask == fullMask,
                                     sigOrVal.updateTime)

    # other conversions should be akreadt done
    return default_auto_cast_fn(self, sigOrVal, toType)



from collections import deque

from hwt.hdl.constants import Time
from hwt.simulator.agentBase import AgentBase, SyncAgentBase
from hwt.synthesizer.exceptions import IntfLvlConfErr


# 100 MHz
DEFAULT_CLOCK = 10 * Time.ns


class SignalAgent(SyncAgentBase):
    """
    Agent for signal interface, it can use clock and reset interface
    for synchronization or can be synchronized by delay

    :attention: clock synchronization has higher priority
    """

    def __init__(self, intf, delay=DEFAULT_CLOCK):
        AgentBase.__init__(self, intf)
        self.delay = delay
        self.initDelay = 0

        # resolve clk and rstn
        try:
            self.clk = self.intf._getAssociatedClk()
        except IntfLvlConfErr:
            self.clk = None

        self._discoverReset(True)
        self.data = deque()

        self.initPending = True

        if self.clk is not None:
            if self.initDelay:
                raise NotImplementedError("initDelay only without clock")
            c = self.SELECTE_EDGE_CALLBACK
            self.monitor = c(self.clk, self.monitor, self.getEnable)
            self.driver = c(self.clk, self.driver, self.getEnable)

    def getDrivers(self):
        d = SyncAgentBase.getDrivers(self)
        if self.clk is None:
            return d
        else:
            return d + [self.driverInit_nonClk]

    def driverInit_nonClk(self, sim):
        try:
            d = self.data[0]
        except IndexError:
            d = None

        self.doWrite(sim, d)

        return
        yield

    def doRead(self, s):
        return s.read(self.intf)

    def doWrite(self, s, data):
        s.write(data, self.intf)

    def driver(self, sim):
        if self.clk is None:
            if self.initPending:
                if self.initDelay:
                    yield sim.wait(self.initDelay)
                self.initPending = False
            # if clock is specified this function is periodically called every
            # clk tick
            while True:
                if self._enabled and self.data and self.notReset(sim):
                    d = self.data.popleft()
                    self.doWrite(sim, d)
                yield sim.wait(self.delay)
        else:
            # if clock is specified this function is periodically called every
            # clk tick, when agent is enabled
            if self.data and self.notReset(sim):
                d = self.data.popleft()
                self.doWrite(sim, d)

    def monitor(self, sim):
        if self.clk is None:
            if self.initPending and self.initDelay:
                yield sim.wait(self.initDelay)
                self.initPending = False
            # if there is no clk, we have to manage periodic call by our selfs
            while True:
                if self._enabled and self.notReset(sim):
                    yield sim.waitOnCombUpdate()
                    d = self.doRead(sim)
                    self.data.append(d)
                    yield sim.wait(self.delay)
        else:
            # if clock is specified this function is periodically called every
            # clk tick, when agent is enabled
            yield sim.waitOnCombUpdate()
            if self.notReset(sim):
                d = self.doRead(sim)
                self.data.append(d)



_indent = "    "
_indentCache = {}


def getIndent(indentNum):
    try:
        return _indentCache[indentNum]
    except KeyError:
        i = "".join([_indent for _ in range(indentNum)])
        _indentCache[indentNum] = i
        return i



from hwt.hdl.types.hdlType import HdlType


class Integer(HdlType):
    """
    Hdl integer type
    """

    def __eq__(self, other):
        return type(self) == type(other)

    def __hash__(self):
        return hash(type(self))

    def all_mask(self):
        return 1

    @classmethod
    def get_auto_cast_fn(cls):
        from hwt.hdl.types.integerCast import cast_integer
        return cast_integer

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.integerVal import IntegerVal
            cls._valCls = IntegerVal
            return cls._valCls



from hwt.hdl.constants import DIRECTION
from hwt.hdl.types.defs import BIT


class NotSpecified(Exception):
    """
    This error means that you need to implement this function to use this functionality

    f.e. you have to implement Simulation agent for interface when you create new one and you can not use existing
    """
    pass


def walkPhysInterfaces(intf):
    if intf._interfaces:
        for si in intf._interfaces:
            yield from walkPhysInterfaces(si)
    else:
        yield intf


def walkParams(intf, discovered):
    """
    walk parameter instances on this interface
    """
    for si in intf._interfaces:
        yield from walkParams(si, discovered)

    for p in intf._params:
        if p not in discovered:
            discovered.add(p)
            yield p


def connectPacked(srcPacked, dstInterface, exclude=None):
    """
    Connect 1D vector signal to this structuralized interface

    :param packedSrc: vector which should be connected
    :param dstInterface: structuralized interface where should
        packedSrc be connected to
    :param exclude: sub interfaces of self which should be excluded
    """
    offset = 0
    connections = []
    for i in reversed(list(walkPhysInterfaces(dstInterface))):
        if exclude is not None and i in exclude:
            continue
        sig = i._sig
        t = sig._dtype
        if t == BIT:
            s = srcPacked[offset]
            offset += 1
        else:
            w = t.bit_length()
            s = srcPacked[(w + offset): offset]
            offset += w
        connections.append(sig(s))

    return connections


def walkFlatten(interface, shouldEnterIntfFn):
    """
    :param shouldEnterIntfFn: function (actual interface)
        returns tuple (shouldEnter, shouldYield)
    """
    _shouldEnter, _shouldYield = shouldEnterIntfFn(interface)
    if _shouldYield:
        yield interface

    if shouldEnterIntfFn:
        for intf in interface._interfaces:
            yield from walkFlatten(intf, shouldEnterIntfFn)
    
        if interface._isInterfaceArray():
            for intf in interface:
                yield from walkFlatten(intf, shouldEnterIntfFn)
    

def packIntf(intf, masterDirEqTo=DIRECTION.OUT, exclude=None):
    """
    Concatenate all signals to one big signal, recursively

    :param masterDirEqTo: only signals with this direction are packed
    :param exclude: sequence of signals/interfaces to exclude
    """
    if not intf._interfaces:
        if intf._masterDir == masterDirEqTo:
            return intf._sig
        return None

    res = None
    for i in intf._interfaces:
        if exclude is not None and i in exclude:
            continue

        if i._interfaces:
            if i._masterDir == DIRECTION.IN:
                d = DIRECTION.opposite(masterDirEqTo)
            else:
                d = masterDirEqTo
            s = i._pack(d, exclude=exclude)
        else:
            if i._masterDir == masterDirEqTo:
                s = i._sig
            else:
                s = None

        if s is not None:
            if res is None:
                res = s
            else:
                res = res._concat(s)

    return res



from hwt.hdl.operator import Operator
from hwt.hdl.portItem import PortItem
from hwt.pyUtils.arrayQuery import arr_any
from hwt.serializer.generic.constants import SIGNAL_TYPE


def _isEventDependentDriver(d):
    return not isinstance(d, (PortItem, Operator)) \
        and d.isEventDependent


def verilogTypeOfSig(signalItem):
    """
    Check if is register or wire
    """
    if signalItem._const or len(signalItem.drivers) > 1 or\
       arr_any(signalItem.drivers, _isEventDependentDriver) or\
       (signalItem._useNopVal
            and len(signalItem.drivers) > 0
            and signalItem.drivers[0].cond):

        return SIGNAL_TYPE.REG
    else:
        return SIGNAL_TYPE.WIRE



from hwt.code import If
from hwt.hdl.architecture import Architecture
from hwt.hdl.assignment import Assignment
from hwt.hdl.entity import Entity
from hwt.hdl.operator import Operator
from hwt.hdl.process import HWProcess
from hwt.hdl.statements import IfContainer, WaitStm, \
    SwitchContainer
from hwt.hdl.types.defs import BIT
from hwt.hdl.value import Value
from hwt.pyUtils.arrayQuery import where, distinctBy, groupedby
from hwt.synthesizer.assigRenderer import renderIfTree
from hwt.synthesizer.exceptions import SigLvlConfErr
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.memory import RtlSyncSignal
from hwt.synthesizer.rtlLevel.optimalizator import removeUnconnectedSignals, \
    reduceProcesses
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from hwt.synthesizer.rtlLevel.signalUtils.walkers import InOutStmProbe
from hwt.synthesizer.rtlLevel.utils import portItemfromSignal


def _isEnclosed(objList):
    if not objList:
        return False
    for o in objList:
        if not isEnclosed(o):
            return False
    return True


def isEnclosed(obj):
    """
    Check if statement has any not used branch
    """
    if isinstance(obj, (Assignment, WaitStm)):
        return True
    elif isinstance(obj, IfContainer):
        for ol in [obj.ifTrue, obj.ifFalse]:
            if not _isEnclosed(ol):
                return False
        for _, ol in obj.elIfs:
            if not _isEnclosed(ol):
                return False

        return True
    elif isinstance(obj, SwitchContainer):
        allCasesCovered = True
        for cond, ol in obj.cases:
            if cond is None:
                allCasesCovered = True
            if not _isEnclosed(ol):
                return False

        return allCasesCovered
    else:
        raise NotImplementedError(obj)


def buildProcessesOutOfAssignments(startsOfDataPaths, getDebugScopeNameFn):
    """
    Render conditional assignments to statements
    and wrap them with process statement
    """
    assigments = where(startsOfDataPaths,
                       lambda x: isinstance(x, Assignment)
                       )
    processes = []
    # process ranks = how many assignments is probably in process
    # used to minimize number of merge tries
    procRanks = {}
    # generate naive processes from assignments
    for sig, dps in groupedby(assigments, lambda x: x.dst):
        dps = list(dps)
        name = ""
        if not sig.hasGenericName:
            name = sig.name
        sig.hidden = False

        haveNotIndexes = True
        for dp in dps:
            haveNotIndexes = haveNotIndexes and not dp.indexes

        # render sequential statements in process
        # (conversion from netlist to statements)
        hasCombDriver = False
        for stm in renderIfTree(dps):
            statements = []
            sProbe = InOutStmProbe()
            sProbe.discover(stm)

            # inject nopVal if needed
            if sig._useNopVal and not isEnclosed(stm):
                n = sig._nopVal
                statements.append(Assignment(n, sig))
                if isinstance(n, RtlSignal):
                    sProbe.sensitivity.add(n)

            statements.append(stm)

            isEventDependent = False
            for s in sProbe.sensitivity:
                if isinstance(s, Operator):
                    # event operator
                    s.operands[0].hidden = False
                    isEventDependent = True
                else:
                    s.hidden = False

            if hasCombDriver and not isEventDependent and haveNotIndexes:
                raise MultipleDriversExc(
                    "%s: Signal %s has multiple combinational drivers" %
                    (getDebugScopeNameFn(), name))

            hasCombDriver = hasCombDriver or not isEventDependent

            outputs = {sig, }
            p = HWProcess("assig_process_" + name,
                          statements, sProbe.sensitivity,
                          sProbe.inputs, outputs)
            processes.append(p)
            procRanks[p] = len(dps)

    yield from reduceProcesses(processes, procRanks)


class RtlNetlist():
    """
    Hierarchical container for signals

    :ivar signals: dict of all signals in context
    :ivar startsOfDataPaths: is set of nodes where datapaths starts
        (assignments)
    :ivar subUnits: is set of all units in this context
    """

    def __init__(self, parentForDebug=None):
        self.parentForDebug = parentForDebug
        self.params = {}
        self.signals = set()
        self.startsOfDataPaths = set()
        self.subUnits = set()
        self.synthesised = False

    def sig(self, name, dtype=BIT, clk=None, syncRst=None, defVal=None):
        """
        generate new signal in context

        :param clk: clk signal, if specified signal is synthesized
            as SyncSignal
        :param syncRst: reset
        """
        if isinstance(defVal, RtlSignal):
            assert defVal._const, \
                "Initial value of register has to be constant"
            _defVal = defVal._auto_cast(dtype)
        elif isinstance(defVal, Value):
            _defVal = defVal._auto_cast(dtype)
        elif isinstance(defVal, InterfaceBase):
            _defVal = defVal._sig
        else:
            _defVal = dtype.fromPy(defVal)

        if clk is not None:
            s = RtlSyncSignal(self, name, dtype, _defVal)
            if syncRst is not None and defVal is None:
                raise Exception(
                    "Probably forgotten default value on sync signal %s", name)
            if syncRst is not None:
                r = If(syncRst._isOn(),
                       [RtlSignal.__call__(s, _defVal)]
                       ).Else(
                    [RtlSignal.__call__(s, s.next)]
                )
            else:
                r = [RtlSignal.__call__(s, s.next)]

            If(clk._onRisingEdge(),
               r
               )
        else:
            if syncRst:
                raise SigLvlConfErr(
                    "Signal %s has reset but has no clk" % name)
            s = RtlSignal(self, name, dtype, defaultVal=_defVal)

        self.signals.add(s)

        return s

    def mergeWith(self, other):
        """
        Merge two instances into this

        :attention: "others" becomes invalid because all signals etc.
            will be transferred into this
        """
        assert not other.synthesised
        self.params.update(other.params)
        self.signals.update(other.signals)
        self.startsOfDataPaths.update(other.startsOfDataPaths)
        self.subUnits.update(other.subUnits)

        for s in other.signals:
            s.ctx = self

    def synthesize(self, name, interfaces):
        """
        Build Entity and Architecture instance out of netlist representation
        """
        ent = Entity(name)
        ent._name = name + "_inst"  # instance name

        # create generics
        for _, v in self.params.items():
            ent.generics.append(v)

        # create ports
        for s in interfaces:
            pi = portItemfromSignal(s, ent)
            pi.reigsterInternSig(s)
            ent.ports.append(pi)

        removeUnconnectedSignals(self)

        # check if all signals are driver by something
        _interfaces = set(interfaces)
        for sig in self.signals:
            if not sig.drivers and sig not in _interfaces:
                assert sig.defaultVal._isFullVld(), (
                    sig, "Signal without any driver or value in ", name)
                sig._const = True

        arch = Architecture(ent)
        for p in buildProcessesOutOfAssignments(self.startsOfDataPaths,
                                                self.getDebugScopeName):
            arch.processes.append(p)

        # add signals, variables etc. in architecture
        for s in self.signals:
            if s.hidden and s.defaultVal.vldMask and not s.drivers:
                # constant
                s.hidden = False

            if s not in interfaces and not s.hidden:
                arch.variables.append(s)

        # instantiate subUnits in architecture
        for u in self.subUnits:
            arch.componentInstances.append(u)

        # add components in architecture
        for su in distinctBy(self.subUnits, lambda x: x.name):
            arch.components.append(su)

        self.synthesised = True

        return [ent, arch]

    def getDebugScopeName(self):
        scope = []
        p = self.parentForDebug
        while p is not None:
            scope.append(p._name)
            try:
                p = p._parent
            except AttributeError:
                break

        return ".".join(reversed(scope))



from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.struct import HStructField, HStruct
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.types.union import HUnion


def HStruct_selectFields(structT, fieldsToUse):
    """
    Select fields from structure (rest will become spacing)

    :param structT: HStruct type instance
    :param fieldsToUse: dict {name:{...}} or set of names to select,
        dictionary is used to select nested fields
        in HStruct or HUnion fields
        (f.e. {"struct1": {"field1", "field2"}, "field3":{}}
        will select field1 and 2 from struct1 and field3 from root)
    """

    template = []
    fieldsToUse = fieldsToUse
    foundNames = set()

    for f in structT.fields:
        name = None
        subfields = []

        if f.name is not None:
            try:
                if isinstance(fieldsToUse, dict):
                    subfields = fieldsToUse[f.name]
                    name = f.name
                else:
                    if f.name in fieldsToUse:
                        name = f.name
            except KeyError:
                name = None

        if name is not None and subfields:
            fields = HStruct_selectFields(f.dtype, subfields)
            template.append(HStructField(fields, name))
        else:
            template.append(HStructField(f.dtype, name))

        if f.name is not None:
            foundNames.add(f.name)

    if isinstance(fieldsToUse, dict):
        fieldsToUse = set(fieldsToUse.keys())
    assert fieldsToUse.issubset(foundNames)

    return HStruct(*template)


def walkFlattenFields(sigOrVal, skipPadding=True):
    """
    Walk all simple values in HStruct or HArray
    """
    t = sigOrVal._dtype
    if isinstance(t, Bits):
        yield sigOrVal
    elif isinstance(t, HUnion):
        yield from walkFlattenFields(sigOrVal._val, skipPadding=skipPadding)
    elif isinstance(t, HStruct):
        for f in t.fields:
            isPadding = f.name is None
            if not isPadding or not skipPadding:
                if isPadding:
                    v = f.dtype.fromPy(None)
                else:
                    v = getattr(sigOrVal, f.name)

                yield from walkFlattenFields(v)

    elif isinstance(t, HArray):
        for item in sigOrVal:
            yield from walkFlattenFields(item)
    else:
        raise NotImplementedError(t)


def HStruct_unpack(structT, data, getDataFn=None, dataWidth=None):
    """
    opposite of packAxiSFrame
    """
    if getDataFn is None:
        assert dataWidth is not None

        def _getDataFn(x):
            return toHVal(x)._auto_cast(Bits(dataWidth))

        getDataFn = _getDataFn

    val = structT.fromPy(None)

    fData = iter(data)

    # actual is storage variable for items from frameData
    actualOffset = 0
    actual = None

    for v in walkFlattenFields(val, skipPadding=False):
        # walk flatten fields and take values from fData and parse them to
        # field
        required = v._dtype.bit_length()

        if actual is None:
            actualOffset = 0
            try:
                actual = getDataFn(next(fData))
            except StopIteration:
                raise Exception("Input data too short")

            if dataWidth is None:
                dataWidth = actual._dtype.bit_length()
            actuallyHave = dataWidth
        else:
            actuallyHave = actual._dtype.bit_length() - actualOffset

        while actuallyHave < required:
            # collect data for this field
            try:
                d = getDataFn(next(fData))
            except StopIteration:
                raise Exception("Input data too short")

            actual = d._concat(actual)
            actuallyHave += dataWidth

        if actuallyHave >= required:
            # parse value of actual to field
            # skip padding
            _v = actual[(required + actualOffset):actualOffset]
            _v = _v._auto_cast(v._dtype)
            v.val = _v.val
            v.vldMask = _v.vldMask
            v.updateTime = _v.updateTime

            # update slice out what was taken
            actuallyHave -= required
            actualOffset += required

        if actuallyHave == 0:
            actual = None

    if actual is not None:
        assert actual._dtype.bit_length(
        ) - actualOffset < dataWidth, "It should be just a padding at the end of frame"

    return val



from copy import copy

from hwt.hdl.assignment import Assignment
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.statements import IfContainer, SwitchContainer, \
    WhileContainer, WaitStm
from hwt.hdl.types.sliceVal import SliceVal
from hwt.hdl.variables import SignalItem
from hwt.pyUtils.arrayQuery import arr_any
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.generic.indent import getIndent
from hwt.serializer.vhdl.utils import VhdlVersion
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from hwt.hdl.types.bits import Bits


class DoesNotContainsTernary(Exception):
    pass


def ternaryOpsToIf(statements):
    """Convert all ternary operators to IfContainers"""
    stms = []

    for st in statements:
        if isinstance(st, Assignment):
            try:
                if not isinstance(st.src, RtlSignalBase):
                    raise DoesNotContainsTernary()
                d = st.src.singleDriver()
                if not isinstance(d, Operator) or d.operator != AllOps.TERNARY:
                    raise DoesNotContainsTernary()
                else:
                    ops = d.operands
                    ifc = IfContainer(ops[0],
                                      [Assignment(ops[1], st.dst)],
                                      [Assignment(ops[2], st.dst)]
                                      )
                    stms.append(ifc)

            except (MultipleDriversExc, DoesNotContainsTernary):
                stms.append(st)
        else:
            stms.append(st)
    return stms


class VhdlSerializer_statements():

    @classmethod
    def Assignment(cls, a, ctx):
        dst = a.dst
        assert isinstance(dst, SignalItem)

        def valAsHdl(v):
            return cls.Value(v, ctx)

        if dst.virtualOnly:
            symbol = ":="
        else:
            symbol = "<="

        if a.indexes is not None:
            for i in a.indexes:
                if isinstance(i, SliceVal):
                    i = i.clone()
                    i.val = (i.val[0], i.val[1])
                dst = dst[i]

        indent_str = getIndent(ctx.indent)
        dstStr = cls.asHdl(dst, ctx)
        src_t = a.src._dtype
        dst_t = dst._dtype

        if dst_t == src_t:
            return "%s%s %s %s" % (indent_str, dstStr, symbol, valAsHdl(a.src))
        else:
            if (isinstance(dst_t, Bits)
                    and isinstance(src_t, Bits)
                    and dst_t.bit_length() == src_t.bit_length() == 1):
                if dst_t.forceVector and not src_t.forceVector:
                    return "%s%s(0) %s %s" % (indent_str, dstStr, symbol,
                                              valAsHdl(a.src))
                if not dst_t.forceVector and src_t.forceVector:
                    return "%s%s %s %s(0)" % (indent_str, dstStr, symbol,
                                              valAsHdl(a.src))

            raise SerializerException(
                "%s%s %s %s  is not valid assignment\n"
                " because types are different (%r; %r) " %
                (indent_str, dstStr, symbol, valAsHdl(a.src),
                 dst._dtype, a.src._dtype))

    @classmethod
    def HWProcess(cls, proc, ctx):
        """
        Serialize HWProcess objects as VHDL

        :param scope: name scope to prevent name collisions
        """
        body = proc.statements
        extraVars = []
        extraVarsSerialized = []

        hasToBeVhdlProcess = arr_any(body,
                                     lambda x: isinstance(x,
                                                          (IfContainer,
                                                           SwitchContainer,
                                                           WhileContainer,
                                                           WaitStm)))

        sensitivityList = sorted(
            map(lambda s: cls.sensitivityListItem(s, None),
                proc.sensitivityList))

        if hasToBeVhdlProcess:
            childCtx = ctx.withIndent()
        else:
            childCtx = copy(ctx)

        def createTmpVarFn(suggestedName, dtype):
            # [TODO] it is better to use RtlSignal
            s = SignalItem(None, dtype, virtualOnly=True)
            s.name = ctx.scope.checkedName(suggestedName, s)
            s.hidden = False
            serializedS = cls.SignalItem(s, childCtx, declaration=True)
            extraVars.append(s)
            extraVarsSerialized.append(serializedS)
            return s

        childCtx.createTmpVarFn = createTmpVarFn

        statemets = [cls.asHdl(s, childCtx) for s in body]
        proc.name = ctx.scope.checkedName(proc.name, proc)

        extraVarsInit = []
        for s in extraVars:
            a = Assignment(s.defaultVal, s, virtualOnly=True)
            extraVarsInit.append(cls.Assignment(a, childCtx))

        _hasToBeVhdlProcess = hasToBeVhdlProcess
        hasToBeVhdlProcess = extraVars or hasToBeVhdlProcess

        if hasToBeVhdlProcess and not _hasToBeVhdlProcess:
            # add indent because we did not added it before because we did not
            # know t
            oneIndent = getIndent(1)
            statemets = list(map(lambda x: oneIndent + x, statemets))

        return cls.processTmpl.render(
            indent=getIndent(ctx.indent),
            name=proc.name,
            hasToBeVhdlProcess=hasToBeVhdlProcess,
            extraVars=extraVarsSerialized,
            sensitivityList=", ".join(sensitivityList),
            statements=extraVarsInit + statemets
        )

    @classmethod
    def IfContainer(cls, ifc, ctx):
        childCtx = ctx.withIndent()

        def asHdl(statements):
            return [cls.asHdl(s, childCtx) for s in statements]

        cond = cls.condAsHdl(ifc.cond, True, childCtx)
        elIfs = []
        if cls.VHDL_VER < VhdlVersion.v2008:
            ifTrue = ternaryOpsToIf(ifc.ifTrue)
            ifFalse = ternaryOpsToIf(ifc.ifFalse)
        else:
            ifTrue = ifc.ifTrue
            ifFalse = ifc.ifFalse

        for c, statements in ifc.elIfs:
            if cls.VHDL_VER < VhdlVersion.v2008:
                statements = ternaryOpsToIf(statements)

            elIfs.append((cls.condAsHdl(c, True, childCtx), asHdl(statements)))

        return cls.ifTmpl.render(
            indent=getIndent(ctx.indent),
            cond=cond,
            ifTrue=asHdl(ifTrue),
            elIfs=elIfs,
            ifFalse=asHdl(ifFalse))

    @classmethod
    def SwitchContainer(cls, sw, ctx):
        childCtx = ctx.withIndent()

        def asHdl(statements):
            return [cls.asHdl(s, childCtx) for s in statements]

        switchOn = cls.condAsHdl(sw.switchOn, False, ctx)

        cases = []
        for key, statements in sw.cases:
            key = cls.asHdl(key, ctx)

            if cls.VHDL_VER < VhdlVersion.v2008:
                statements = ternaryOpsToIf(statements)

            cases.append((key, asHdl(statements)))

        if sw.default:
            cases.append((None, asHdl(sw.default)))

        return cls.switchTmpl.render(
            indent=getIndent(ctx.indent),
            switchOn=switchOn,
            cases=cases)

    @classmethod
    def WaitStm(cls, w, ctx):
        indent_str = getIndent(ctx.indent)

        if w.isTimeWait:
            return "%swait for %d ns" % (indent_str, w.waitForWhat)
        elif w.waitForWhat is None:
            return "%swait" % indent_str
        else:
            raise NotImplementedError()



"""
AssigRenderer is responsible for converting sequence of conditional statements
(usually assignments) to code elements like if-then-else statements
to realize conditions of original statements

List of conditional statements is used to simplify code manipulation.

:var SWITCH_THRESHOLD: max count of elsifs with eq on same variable
    to convert this if-then-else statement to switch statement
"""

from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.statements import IfContainer, SwitchContainer
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.enum import HEnum
from hwt.hdl.value import Value
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from hwt.synthesizer.termUsageResolver import getBaseCond


SWITCH_THRESHOLD = 2


def condWithoutResolved(cond, resolvedCnt):
    """
    Filter already resolved items from condition list of statement
    """
    return reversed(cond[:len(cond) - resolvedCnt])


def splitStatementsOnCond(statements, resolvedCondCnt):
    """
    Try to discover condition for top if statement

    :param statements: sequence of statements
    :param resolvedCnt: number of resolved conditions in concition sets
        of statements
    """

    # resolve how many condition items can we take into into actual if
    # statement
    simplestStm = statements[0]
    cntOfSameConditions = len(simplestStm.cond)
    for a in statements:
        _cntOfSameConditions = 0
        for c0, c1 in zip(condWithoutResolved(a.cond, resolvedCondCnt),
                          condWithoutResolved(simplestStm.cond,
                                              resolvedCondCnt)):
            if c0 is c1:
                _cntOfSameConditions += 1
            else:
                break

        if _cntOfSameConditions < cntOfSameConditions:
            cntOfSameConditions = _cntOfSameConditions

        if len(a.cond) < len(simplestStm.cond):
            simplestStm = a

    assert cntOfSameConditions >= 0
    ifFalse = []
    independent = []
    if cntOfSameConditions > 0:
        # if cond:
        #   stm
        condLen = len(simplestStm.cond)
        high = condLen - resolvedCondCnt - cntOfSameConditions
        low = condLen - resolvedCondCnt
        topCond = simplestStm.cond[high:low]
        ifTrue = list(renderIfTree(
            statements, resolvedCondCnt + cntOfSameConditions))
    else:
        # if cond:
        #    stm
        # else:
        #    stm
        condIndx = -resolvedCondCnt - 1
        try:
            _topCond = simplestStm.cond[condIndx]
        except IndexError:
            raise Exception(
                "Error while resolving position in if-tree of statement %r"
                " with conditions %s" % (simplestStm, simplestStm.cond))

        topCond, _ = getBaseCond(_topCond)
        topCondNeg = ~topCond

        ifTrue = []
        for s in statements:
            c = s.cond[condIndx]
            if c is topCond:
                ifTrue.append(s)
            elif c is topCondNeg:
                ifFalse.append(s)
            else:
                independent.append(s)

        ifTrue = list(renderIfTree(ifTrue, resolvedCondCnt + 1))
        ifFalse = list(renderIfTree(ifFalse, resolvedCondCnt + 1))
        topCond = [topCond, ]

    assert topCond
    return topCond, ifTrue, ifFalse, independent


def withoutItemOnIndex(items, index):
    """
    :return: generator of items from "items" without item
        on specified index
    """
    for i, item in enumerate(items):
        if i != index:
            yield item


def _disolveConditionAsEq(cond):
    try:
        op = cond.singleDriver()
    except MultipleDriversExc:
        return
    if isinstance(op, Operator) and op.operator == AllOps.EQ:
        op0, op1 = op.operands
        if isinstance(op0, RtlSignalBase) and isinstance(op1, Value):
            return op0, op1
        elif isinstance(op1, RtlSignalBase) and isinstance(op0, Value):
            return op1, op0


def disolveConditionAsEq(condList):
    """
    detect if condition is in format signal == value

    :return: None if condList is not in format [ a == val] or [val == a]
        else return tuple (a, val, restOfCodList)
    """
    rest = []
    if len(condList) == 1:
        dis = _disolveConditionAsEq(condList[0])
    else:
        for i, cond in enumerate(condList):
            dis = _disolveConditionAsEq(cond)
            if dis is not None:
                rest = list(withoutItemOnIndex(condList, i))
                break

    if dis is not None:
        a, val = dis
        return (a, val, rest)


def typeDomainSize(t):
    """
    :return: how many values can have specified type
    """
    if isinstance(t, (HEnum, Bits)):
        return 2 ** t.bit_length()
    else:
        raise TypeError(t)


def renderIfTree(statements, resolvedCnt=0):
    """
    Walk assignments and resolve if tree from conditions

    :param statements: sequence of statements
    :param resolvedCnt: number of resolved conditions in concition sets
        of statements
    """

    # filter statements which are not under any condition
    _statements = []
    for a in statements:
        if len(a.cond) > resolvedCnt:
            _statements.append(a)
        else:
            yield a

    if _statements:
        (topCond,
         ifTrue, ifFalse,
         independent) = splitStatementsOnCond(_statements, resolvedCnt)
        if independent:
            yield from renderIfTree(independent, resolvedCnt)

        # reduce if else to elifs or switch
        elIfs = []
        while len(ifFalse) == 1:
            stm = ifFalse[0]
            if isinstance(stm, IfContainer):
                elIfs.extend(stm.elIfs)
                elIfs.insert(0, (stm.cond, stm.ifTrue))
                ifFalse = stm.ifFalse
                continue
            elif isinstance(stm, SwitchContainer):
                # try to extend switch statement
                dis = disolveConditionAsEq(topCond)
                if dis:
                    switchOn, topVal, restOfCond = dis
                    if stm.switchOn is switchOn:
                        assert not elIfs
                        if restOfCond:
                            ifTrue = IfContainer.potentialyReduced(restOfCond,
                                                                   ifTrue)
                        stm.cases.insert(0, (topVal, ifTrue))

                        t = switchOn._dtype
                        # it type is enum and all values are used in switch and
                        # if enum bit representation can have more values than
                        # enum itself try to set last case as default
                        # to prevent latches in hdl
                        if isinstance(t, HEnum):
                            setDefault = False
                            try:
                                tValues = typeDomainSize(t)
                                caseLen = len(stm.cases)
                                setDefault = (not stm.default
                                              and tValues != caseLen)
                            except TypeError:
                                pass
                            if setDefault:
                                # we can use any item as default
                                # because switch contains case for
                                # every possible value from type
                                c = stm.cases.pop()
                                stm.default = c[1]
                        yield stm
                        return
            break

        if len(elIfs) >= SWITCH_THRESHOLD:
            # try to convert elifs to switch
            cases = []
            disolvedTopCond = disolveConditionAsEq(topCond)
            if disolvedTopCond is not None:
                switchOn, topVal, restOfCond = disolvedTopCond
                if restOfCond:
                    ifTrue = IfContainer.potentialyReduced(restOfCond,
                                                           ifTrue)
                cases.append((topVal, ifTrue))
                canBeConvertedToSwitch = True
                for elIf in elIfs:
                    dis = disolveConditionAsEq(elIf[0])
                    if dis is None or dis[0] is not switchOn:
                        canBeConvertedToSwitch = False
                        break
                    else:
                        _, v, restOfCond = dis
                        stms = elIf[1]
                        if restOfCond:
                            stms = IfContainer.potentialyReduced(restOfCond,
                                                                 stms)
                        cases.append((v, stms))

                # if only last can not be part of the switch case it can be
                # default
                if not canBeConvertedToSwitch and len(elIfs) == len(cases):
                    default = elIfs[-1]
                    ifFalse = IfContainer.potentialyReduced(default[0],
                                                            default[1],
                                                            ifFalse)
                    canBeConvertedToSwitch = True

                if canBeConvertedToSwitch:
                    t = switchOn._dtype
                    if not ifFalse:
                        # it type is enum and all values are used in switch and
                        # if enum bit representation can have more values than
                        # enum itself try to set last case as default
                        # to prevent latches in hdl
                        if (isinstance(t, HEnum) and
                                typeDomainSize(t) > len(t._allValues) and
                                len(t._allValues) == len(cases)):
                            ifFalse = cases[-1][1]
                            cases = cases[:-1]
                    default = ifFalse

                    yield SwitchContainer(switchOn, cases, default)
                    return

        # try to reduce redundant "if"s
        if not ifFalse and not elIfs:
            # try reduce
            # if a:
            #     if b:
            # to if a and b:
            try:
                subIf, = ifTrue
            except ValueError:
                subIf = None

            if (isinstance(subIf, IfContainer) and
                    subIf.ifTrue and
                    not subIf.ifFalse and
                    not subIf.elIfs):
                topCond = list(reversed(topCond))
                topCond.extend(subIf.cond)
                subIf.cond = topCond
                yield subIf
                return

        yield from IfContainer.potentialyReduced(
            list(reversed(topCond)),
            ifTrue,
            ifFalse,
            elIfs)



from hwt.hdl.variables import SignalItem


class SimSignal(SignalItem):
    """
    Class of signal simulation functions

    :ivar _writeCallbacks: list of callback functions(signal, simulator)
        which is called when new (changed) value is written to this signal
    """
    __slots__ = ["name", "_val", "_oldVal", "_writeCallbacks",
                 "simSensProcs", "simRisingSensProcs", "simFallingSensProcs"]

    def __init__(self, ctx, name, dtype, defaultVal=None):
        ctx.signals.add(self)
        self.hidden = False
        self._writeCallbacks = []
        self._writeCallbacksToEn = []
        self.simSensProcs = set()
        self.simRisingSensProcs = set()
        self.simFallingSensProcs = set()
        super(SimSignal, self).__init__(name, dtype, defaultVal)

    def registerWriteCallback(self, callback, getEnFn) -> int:
        """
        Register writeCallback for signal.
        Registration is evaluated at the end of deltastep of simulator.

        :param callback: simulation process represented by function(simulator)
            which should be called after update of this signal
        :param getEnFn: function() to get initial value for enable of callback
        """
        index = len(self._writeCallbacks)
        self._writeCallbacks.append(None)
        self._writeCallbacksToEn.append((index, callback, getEnFn))
        return index

    def _loadWriteCallbacks(self):
        wc = self._writeCallbacksToEn
        self._writeCallbacksToEn = []
        # perform registration of new write callbacks
        for i, callback, reqEnFn in wc:
            if reqEnFn():
                self._writeCallbacks[i] = callback

    def simPropagateChanges(self, simulator):
        v = self._val
        self._oldVal = v

        if self._writeCallbacksToEn:
            self._loadWriteCallbacks()
        # run all sensitive processes
        log = simulator.config.logPropagation
        if log:
            log(simulator, self, self.simSensProcs)
        for p in self.simSensProcs:
            simulator.addHwProcToRun(self, p)

        # run write callbacks we have to create new list to allow
        # registering of new call backs in callbacks
        for c in self._writeCallbacks:
            if c:
                # run simulation processes which are activated
                simulator.process(c(simulator))

        if self.simRisingSensProcs:
            if v.val or not v.vldMask:
                if log:
                    log(simulator, self, self.simRisingSensProcs)
                for p in self.simRisingSensProcs:

                    simulator.addHwProcToRun(self, p)

        if self.simFallingSensProcs:
            if not v.val or not v.vldMask:
                if log:
                    log(simulator, self, self.simFallingSensProcs)
                for p in self.simFallingSensProcs:
                    simulator.addHwProcToRun(self, p)

    def simUpdateVal(self, simulator, valUpdater):
        """
        Method called by simulator to update new value for this object
        """
        dirtyFlag, newVal = valUpdater(self._oldVal)

        if dirtyFlag:
            self._val = newVal
            newVal.updateTime = simulator.now
            log = simulator.config.logChange
            if log:
                log(simulator.now, self, newVal)

            self.simPropagateChanges(simulator)



from hwt.code import log2ceil
from hwt.hdl.constants import DIRECTION
from hwt.interfaces.agents.unionIntf import UnionSourceAgent
from hwt.interfaces.std import Handshaked
from hwt.interfaces.structIntf import StructIntf


class UnionSink(StructIntf):
    """
    Interface generated from HUnion HDL type

    Used when consumer chooses which member of union should be used.
    """

    def _declr(self):
        StructIntf._declr(self)
        self._select = Handshaked()
        self._select.DATA_WIDTH.set(log2ceil(len(self._structT.fields)))


class UnionSource(UnionSink):
    """
    Same like `UnionSink` but producer is selecting member of union
    which should be used.
    """

    def _declr(self):
        StructIntf._declr(self)
        self._select = Handshaked(masterDir=DIRECTION.IN)
        self._select.DATA_WIDTH.set(log2ceil(len(self._structT.fields)))

    def _initSimAgent(self):
        self._ag = UnionSourceAgent(self)



from hwt.interfaces.agents.structIntf import StructIntfAgent, getDrivers, \
    getMonitors


class UnionSourceAgent(StructIntfAgent):
    def getMonitors(self):
        sel = self.intf._select
        for intf in self.intf._interfaces:
            if intf is sel:
                yield from getDrivers(intf)
            else:
                yield from getMonitors(intf)

    def getDrivers(self):
        sel = self.intf._select
        for intf in self.intf._interfaces:
            if intf is sel:
                yield from getMonitors(intf)
            else:
                yield from getDrivers(intf)



from hwt.hdl.types.array import HArray
from hwt.hdl.types.integer import Integer


class Slice(HArray):
    """
    Slice type, used for selecting items from arrays or vectors
    """

    def __init__(self):
        super().__init__(Integer(), 2)

    def __hash__(self):
        return hash(self.elmType)

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.sliceVal import SliceVal
            cls._valCls = SliceVal
            return cls._valCls



from hwt.hdl.hdlObject import HdlObject
from hwt.synthesizer.andReducedContainer import AndReducedContainer


class Assignment(HdlObject):
    """
    Assignment container

    :ivar src: source
    :ivar dest: destination signal
    :ivar cond: set of terms if all them are evaluated to True,
        assignment is active
    :ivar indexes: description of index selector on dst
        (list of Index/Slice objects) (f.e. [[0], [1]] means  dst[0][1])

    :cvar __instCntr: counter used for generating instance ids
    :ivar _instId: internaly used only for intuitive sorting of statements

    """
    __instCntr = 0

    def __init__(self, src, dst, indexes=None, virtualOnly=False):
        """
        :param dst: destination to assign to
        :param src: source which is assigned from
        :param indexes: description of index selector on dst
            (list of Index/Slice objects) (f.e. [[0], [1]] means  dst[0][1])
        :param virtualOnly: flag indicates that this assignments
            is only virtual and should not be added into
            netlist, because it is only for internal notation
        """
        self.src = src
        self.dst = dst
        self.isEventDependent = False
        self.indexes = indexes
        self.cond = AndReducedContainer()
        self._instId = Assignment._nextInstId()

        if not virtualOnly:
            dst.ctx.startsOfDataPaths.add(self)

    @classmethod
    def _nextInstId(cls):
        """
        Get next instance id
        """
        i = cls.__instCntr
        cls.__instCntr += 1
        return i

    def seqEval(self):
        """Sequentially evaluate this assignment"""
        self.dst._val = self.src.staticEval()



from hwt.synthesizer.exceptions import TypeConversionErr


def default_reinterpret_cast_fn(typeFrom, sigOrVal, toType):
    raise TypeConversionErr(
        "Reinterpretation of %r of type \n%r to type %r is not implemented",
        (sigOrVal, typeFrom, toType))


def default_auto_cast_fn(typeFrom, sigOrVal, toType):
    raise TypeConversionErr(
        "Conversion of %r of type \n%r to type %r is not implemented",
        (sigOrVal, typeFrom, toType))


class HdlType():
    """
    Base class for all hardware related types.

    :ivar _auto_cast_fn: convert function (attribute set on first convert
        function call)
    :ivar _reinterpret_cast_fn: reinterpret function (attribute set
        on first convert function call)
    """

    def fromPy(self, v, vldMask=None):
        """
        Construct value of this type.
        Delegated on value class for this type
        """
        return self.getValueCls().fromPy(v, self, vldMask=vldMask)

    def auto_cast(self, sigOrVal, toType):
        """
        Cast value or signal of this type to another compatible type.

        :param sigOrVal: instance of signal or value to cast
        :param toType: instance of HdlType to cast into
        """
        if sigOrVal._dtype == toType:
            return sigOrVal

        try:
            c = self._auto_cast_fn
        except AttributeError:
            c = self.get_auto_cast_fn()
            self._auto_cast_fn = c

        return c(self, sigOrVal, toType)

    def reinterpret_cast(self, sigOrVal, toType):
        """
        Cast value or signal of this type to another type of same size.

        :param sigOrVal: instance of signal or value to cast
        :param toType: instance of HdlType to cast into
        """
        try:
            return self.auto_cast(sigOrVal, toType)
        except TypeConversionErr:
            pass

        try:
            r = self._reinterpret_cast_fn
        except AttributeError:
            r = self.get_reinterpret_cast_fn()
            self._reinterpret_cast_fn = r

        return r(self, sigOrVal, toType)

    @classmethod
    def get_auto_cast_fn(cls):
        """
        Get method for converting type
        """
        return default_auto_cast_fn

    @classmethod
    def get_reinterpret_cast_fn(cls):
        """
        Get method for converting type
        """
        return default_reinterpret_cast_fn

    @classmethod
    def getValueCls(cls):
        """
        :attention: Overrode in implementation of concrete HdlType.

        :return: class for value derived from this type
        """
        raise NotImplementedError()

    def __getitem__(self, key):
        """
        [] operator to create an array of this type.
        """
        assert int(key) > 0, key  # array has to have some items
        from hwt.hdl.types.array import HArray
        return HArray(self, key)

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not None is used as a additional
            information about on which address this type is stored
            (used only by HStruct)
        :param expandStructs: expand HStructTypes (used by HStruct and Array)
        """
        return "<%s>" % (self.__class__.__name__)





SYSTEMC_KEYWORDS = [
    # c++11 keywords
    "alignas",
    "alignof",
    "and",
    "and_eq",
    "asm",
    "atomic_cancel",
    "atomic_commit",
    "atomic_noexcept",
    "auto",
    "bitand",
    "bitor",
    "bool",
    "break",
    "case",
    "catch",
    "char",
    "char16_t",
    "char32_t",
    "class",
    "compl",
    "concept",
    "const",
    "constexpr",
    "const_cast",
    "continue",
    "decltype",
    "default",
    "delete",
    "do",
    "double",
    "dynamic_cast",
    "else",
    "enum",
    "explicit",
    "export",
    "extern",
    "false",
    "float",
    "for",
    "friend",
    "goto",
    "if",
    "import",
    "inline",
    "int",
    "long",
    "module",
    "mutable",
    "namespace",
    "new",
    "noexcept",
    "not",
    "not_eq",
    "nullptr",
    "operator",
    "or",
    "or_eq",
    "private",
    "protected",
    "public",
    "register",
    "reinterpret_cast",
    "requires",
    "return",
    "short",
    "signed",
    "sizeof",
    "static",
    "static_assert",
    "static_cast",
    "struct",
    "switch",
    "synchronized",
    "template",
    "this",
    "thread_local",
    "throw",
    "true",
    "try",
    "typedef",
    "typeid",
    "typename",
    "union",
    "unsigned",
    "using",
    "virtual",
    "void",
    "volatile",
    "wchar_t",
    "while",
    "xor",
    "xor_eq",
    # SystemC namespace
    "SC_MODULE", "SC_METHOD", "SC_THREAD", "SC_CTHREAD", "SC_CTOR", "SC_BIN",
    "SC_OCT", "SC_HEX", "SC_DEC",
    "sensitive",
    "sc_signal", "sc_in", "sc_out", "sc_inout", "sc_event", "sc_bit",
    "sc_fixed", "sc_fixed", "sc_lv", "sc_bv", "sc_logic",
    "sc_bigint", "sc_biguint",  "sc_int", "sc_out_resolved",
    "sc_inout_resolve", "sc_out_rv",
    "sc_inout_rv", "sc_signal_resolved", "sc_signal_rv",
    "notify", "sc_module", "sc_inout_clk", "sc_out_clk",
    "sc_create_vcd_trace_file", "sc_trace_file", "sc_trace"]




from jinja2.environment import Environment
from jinja2.loaders import PackageLoader


class VhdlTmplContainer():
    tmplEnv = Environment(loader=PackageLoader(
        'hwt', 'serializer/vhdl/templates'))
    architectureTmpl = tmplEnv.get_template('architecture.vhd')
    entityTmpl = tmplEnv.get_template('entity.vhd')
    processTmpl = tmplEnv.get_template('process.vhd')
    componentTmpl = tmplEnv.get_template('component.vhd')
    componentInstanceTmpl = tmplEnv.get_template('component_instance.vhd')
    ifTmpl = tmplEnv.get_template('if.vhd')
    switchTmpl = tmplEnv.get_template('switch.vhd')



from hwt.hdl.constants import Time
from hwt.simulator.agentBase import AgentBase


def pullDownAfter(sig, intDelay=6 * Time.ns):
    """
    :return: simulation driver which keeps signal value high for intDelay
        then it sets value to 0
    """
    def _pullDownAfter(s):
        s.write(True, sig)
        yield s.wait(intDelay)
        s.write(False, sig)

    return _pullDownAfter


def pullUpAfter(sig, intDelay=6 * Time.ns):
    """
    :return: Simulation driver which keeps signal value low for intDelay then
        it sets value to 1
    """
    def _pullDownAfter(s):
        s.write(False, sig)
        yield s.wait(intDelay)
        s.write(True, sig)

    return _pullDownAfter


class PullUpAgent(AgentBase):
    def __init__(self, intf, intDelay=6 * Time.ns):
        self.intDelay = intDelay
        self.data = []
        self.driver = pullUpAfter(intf, intDelay=intDelay)


class PullDownAgent(AgentBase):
    def __init__(self, intf, intDelay=6 * Time.ns):
        self.intDelay = intDelay
        self.data = []
        self.driver = pullDownAfter(intf, intDelay=intDelay)



from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.hdl.operatorDefs import AllOps


def isResultOfTypeConversion(sig):
    try:
        sig.drivers[0]
    except IndexError:
        return False

    if sig.hidden:
        return True
    return False


class VhdlSerializer_ops():
    # keep in mind that there is no such a thing in vhdl itself
    opPrecedence = {
        AllOps.NOT: 2,
        AllOps.RISING_EDGE: 1,
        AllOps.NEG: 2,
        AllOps.DIV: 3,
        AllOps.ADD: 3,
        AllOps.SUB: 3,
        AllOps.MUL: 3,
        AllOps.XOR: 2,
        AllOps.EQ: 2,
        AllOps.NEQ: 2,
        AllOps.AND: 2,
        AllOps.OR: 2,
        AllOps.DOWNTO: 2,
        AllOps.GT: 2,
        AllOps.LT: 2,
        AllOps.GE: 2,
        AllOps.LE: 2,
        AllOps.CONCAT: 2,
        AllOps.INDEX: 1,
        AllOps.TERNARY: 1,
        AllOps.CALL: 1,
    }
    _binOps = {
        AllOps.AND: '%s AND %s',
        AllOps.OR: '%s OR %s',
        AllOps.XOR: '%s XOR %s',
        AllOps.CONCAT: '%s & %s',
        AllOps.DIV: '%s / %s',
        AllOps.DOWNTO: '%s-1 DOWNTO %s',
        AllOps.TO: '%s-1 TO %s',
        AllOps.EQ: '%s = %s',
        AllOps.GT: '%s > %s',
        AllOps.GE: '%s >= %s',
        AllOps.LE: '%s <= %s',
        AllOps.POW: '%s ** %s',
        AllOps.LT: '%s < %s',
        AllOps.SUB: '%s - %s',
        AllOps.MUL: '%s * %s',
        AllOps.NEQ: '%s /= %s',
        AllOps.ADD: '%s + %s',
    }
    _unaryOps = {
        AllOps.NOT: "NOT %s",
        AllOps.NEG: "-(%s)",
        AllOps.RISING_EDGE: "RISING_EDGE(%s)",
        AllOps.FALLIGN_EDGE: "FALLING_EDGE(%s)",
        AllOps.BitsAsSigned: "SIGNED(%s)",
        AllOps.BitsAsUnsigned: "UNSIGNED(%s)",
        AllOps.BitsAsVec: "STD_LOGIC_VECTOR(%s)",
    }

    @classmethod
    def Operator(cls, op, ctx):
        # [TODO] no nested ternary in expressions like
        # ( '1'  WHEN r = f ELSE  '0' ) & "0"
        ops = op.operands
        o = op.operator

        op_str = cls._unaryOps.get(o, None)
        if op_str is not None:
            return op_str % (cls._operand(ops[0], o, ctx))

        op_str = cls._binOps.get(o, None)
        if op_str is not None:
            return op_str % (cls._operand(ops[0], o, ctx), cls._operand(ops[1], o, ctx))

        if o == AllOps.CALL:
            return "%s(%s)" % (cls.FunctionContainer(ops[0]),
                               ", ".join(map(lambda op: cls._operand(op, o, ctx), ops[1:])))
        elif o == AllOps.INDEX:
            assert len(ops) == 2
            o1 = ops[0]
            if isinstance(o1, RtlSignalBase) and isResultOfTypeConversion(o1):
                o1 = ctx.createTmpVarFn("tmpTypeConv", o1._dtype)
                o1.defaultVal = ops[0]

            return "%s(%s)" % (cls.asHdl(o1, ctx).strip(), cls._operand(ops[1], o, ctx))
        elif o == AllOps.TERNARY:
            return " ".join([cls._operand(ops[1], o, ctx), "WHEN",
                             cls.condAsHdl([ops[0]], True, ctx),
                             "ELSE",
                             cls._operand(ops[2], o, ctx)])
        elif o == AllOps.BitsToInt:
            assert len(ops) == 1
            op = cls.asHdl(ops[0], ctx)
            if ops[0]._dtype.signed is None:
                op = "UNSIGNED(%s)" % op
            return "TO_INTEGER(%s)" % op
        elif o == AllOps.IntToBits:
            assert len(ops) == 1
            resT = op.result._dtype
            op_str = cls.asHdl(ops[0], ctx)
            w = resT.bit_length()

            if resT.signed is None:
                return "STD_LOGIC_VECTOR(TO_UNSIGNED(%s, %d))" % (op_str, w)
            elif resT.signed:
                return "TO_UNSIGNED(%s, %d)" % (op_str, w)
            else:
                return "TO_UNSIGNED(%s, %d)" % (op_str, w)
        else:
            raise NotImplementedError(
                "Do not know how to convert %s to vhdl" % (o))



from hwt.hdl.constants import DIRECTION
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BIT, BIT_N
from hwt.interfaces.agents.bramPort import BramPortAgent
from hwt.interfaces.agents.bramPort import BramPort_withoutClkAgent
from hwt.interfaces.agents.clk import OscilatorAgent
from hwt.interfaces.agents.fifo import FifoReaderAgent
from hwt.interfaces.agents.fifo import FifoWriterAgent
from hwt.interfaces.agents.handshaked import HandshakeSyncAgent
from hwt.interfaces.agents.handshaked import HandshakedAgent
from hwt.interfaces.agents.rdSynced import RdSyncedAgent
from hwt.interfaces.agents.regCntrl import RegCntrlAgent
from hwt.interfaces.agents.rst import PullDownAgent
from hwt.interfaces.agents.rst import PullUpAgent
from hwt.interfaces.agents.signal import SignalAgent
from hwt.interfaces.agents.vldSynced import VldSyncedAgent
from hwt.interfaces.signalOps import SignalOps
from hwt.synthesizer.interface import Interface
from hwt.synthesizer.param import Param
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


D = DIRECTION


class Signal(SignalOps, Interface):
    """
    Basic wire interface
    """

    def __init__(self,
                 masterDir=D.OUT,
                 asArraySize=None,
                 dtype=BIT,
                 loadConfig=True):
        super().__init__(masterDir=masterDir,
                         asArraySize=asArraySize,
                         loadConfig=loadConfig)
        self._dtype = dtype

    def _injectMultiplerToDtype(self):
        """
        Make signal wider, used when there is an array of signals stored
        in one wider signal
        """
        t = self._dtype
        factor = self._widthMultiplier
        if t == BIT:
            newT = Bits(factor)
        elif isinstance(t, Bits):
            w = t.width
            if isinstance(w, int):
                newW = factor * w
            elif isinstance(w, RtlSignalBase):
                # both Param or factor Value
                newW = w * factor
            elif isinstance(factor, RtlSignalBase):
                # w is Value
                newW = factor * w
            else:
                # both Value
                newW = w.clone()
                newW.val *= factor.val
            newT = Bits(newW)
        else:
            raise TypeError("Can not multiply width of type %r" % (repr(t),))

        self._dtype = newT

    def _initSimAgent(self):
        self._ag = SignalAgent(self)


def VectSignal(width,
               signed=None,
               masterDir=D.OUT,
               asArraySize=None,
               loadConfig=True):
    """
    Create basic :class:`.Signal` interface where type is vector
    """
    return Signal(masterDir,
                  asArraySize,
                  Bits(width, signed, forceVector=True),
                  loadConfig)


class Clk(Signal):
    """
    Basic :class:`.Signal` interface which is interpreted as clock signal
    """

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Clk
        return IP_Clk

    def _initSimAgent(self):
        self._ag = OscilatorAgent(self)


class Rst(Signal):
    """
    Basic :class:`.Signal` interface which is interpreted as reset signal
    """

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Rst
        return IP_Rst

    def _initSimAgent(self):
        self._ag = PullDownAgent(self)


class Rst_n(Signal):
    """
    Basic :class:`.Signal` interface which is interpreted as reset signal
    with negative polarity (active in 0)
    """

    def __init__(self,
                 masterDir=D.OUT,
                 asArraySize=None,
                 dtype=BIT_N,
                 loadConfig=True):
        super(Rst_n, self).__init__(masterDir=D.OUT,
                                    asArraySize=asArraySize,
                                    dtype=dtype,
                                    loadConfig=dtype)

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Rst_n
        return IP_Rst_n

    def _initSimAgent(self):
        self._ag = PullUpAgent(self)


class VldSynced(Interface):
    """
    Interface data+valid signal, if vld=1 then data are valid and slave should
    accept them
    """

    def _config(self):
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.data = VectSignal(self.DATA_WIDTH)
        self.vld = s()

    def _initSimAgent(self):
        self._ag = VldSyncedAgent(self)


class RdSynced(Interface):
    """
    Interface data+ready signal, if rd=1 then slave has read data and master
    should actualize data
    """

    def _config(self):
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.data = VectSignal(self.DATA_WIDTH)
        self.rd = s(masterDir=D.IN)

    def _initSimAgent(self):
        self._ag = RdSyncedAgent(self)


class Handshaked(VldSynced):
    """
    Interface data+ready+valid signal, if rd=1 slave is ready to accept data,
    if vld=1 master is sending data,
    if rd=1 and vld=1 then data is transfered otherwise master
    and slave has to wait on each other

    :attention: one rd/vld is set it must not go down until transaction is made
    """

    def _declr(self):
        super()._declr()
        self.rd = s(masterDir=D.IN)

    def _initSimAgent(self):
        self._ag = HandshakedAgent(self)


class HandshakeSync(Interface):
    """
    Only synchronization interface, like vld+rd signal with meaning
    like in :class:`.Handshaked` interface

    :ivar rd: when high slave is ready to receive data
    :ivar vld: when high master is sending data to slave

    transaction happens when both ready and valid are high

    """

    def _declr(self):
        self.vld = s()
        self.rd = s(masterDir=D.IN)

    def _initSimAgent(self):
        self._ag = HandshakeSyncAgent(self)


class ReqDoneSync(Interface):
    """
    Synchronization interface, if req=1 slave begins operation and when
    it's done it asserts done=1 for one clk tick req does not need to stay high
    """

    def _declr(self):
        self.req = s()
        self.done = s(masterDir=D.IN)


class BramPort_withoutClk(Interface):
    """
    Basic BRAM port
    """

    def _config(self):
        self.ADDR_WIDTH = Param(32)
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.addr = VectSignal(self.ADDR_WIDTH)
        self.din = VectSignal(self.DATA_WIDTH)
        self.dout = VectSignal(self.DATA_WIDTH, masterDir=D.IN)
        self.en = s()
        self.we = s()

    def _getWordAddrStep(self):
        """
        :return: size of one word in unit of address
        """
        return 1

    def _getAddrStep(self):
        """
        :return: how many bits is one unit of address (f.e. 8 bits for
            char * pointer, 36 for 36 bit bram)
        """
        return int(self.DATA_WIDTH)

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_BlockRamPort
        return IP_BlockRamPort

    def _initSimAgent(self):
        self._ag = BramPort_withoutClkAgent(self)


class BramPort(BramPort_withoutClk):
    """
    BRAM port with it's own clk
    """

    def _declr(self):
        self.clk = s(masterDir=D.OUT)
        with self._associated(clk=self.clk):
            super()._declr()

        self._associatedClk = self.clk

    def _initSimAgent(self):
        self._ag = BramPortAgent(self)


class FifoWriter(Interface):
    def _config(self):
        self.DATA_WIDTH = Param(8)

    def _declr(self):
        self.en = s()
        self.wait = s(masterDir=D.IN)
        self.data = VectSignal(self.DATA_WIDTH)

    def _initSimAgent(self):
        self._ag = FifoWriterAgent(self)


class FifoReader(FifoWriter):
    def _declr(self):
        super()._declr()
        self.en._masterDir = D.IN
        self.wait._masterDir = D.OUT

    def _initSimAgent(self):
        self._ag = FifoReaderAgent(self)


class RegCntrl(Interface):
    """
    Register control interface, :class:`.Signal` for read, :class:`.VldSynced`
    for write
    """

    def _config(self):
        self.DATA_WIDTH = Param(8)

    def _declr(self):
        self.din = VectSignal(self.DATA_WIDTH, masterDir=D.IN)
        with self._paramsShared():
            self.dout = VldSynced()

    def _initSimAgent(self):
        self._ag = RegCntrlAgent(self)


s = Signal



from collections import deque

from hwt.simulator.agentBase import SyncAgentBase
from hwt.simulator.shortcuts import OnRisingCallbackLoop
from hwt.interfaces.agents.signal import DEFAULT_CLOCK


class FifoReaderAgent(SyncAgentBase):
    """
    Simulation agent for FifoReader interface
    """

    def __init__(self, intf, allowNoReset=False):
        super(FifoReaderAgent, self).__init__(intf, allowNoReset)
        self.data = deque()
        self.readPending = False
        self.lastData = None

        # flags to keep data coherent when enable state changes
        self.lastData_invalidate = False
        self.readPending_invalidate = False

    def setEnable_asDriver(self, en, sim):
        self._enabled = en
        self.driver.setEnable(en, sim)
        sim.write(not en, self.intf.wait)
        self.lastData_invalidate = not en

    def setEnable_asMonitor(self, en, sim):
        lastEn = self._enabled
        self._enabled = en
        self.monitor.setEnable(en, sim)
        sim.write(en, self.intf.en)
        self.readPending_invalidate = not en
        if not lastEn:
            self.dataReader.setEnable(en, sim)

    def driver_init(self, sim):
        sim.write(not self._enabled, self.intf.wait)
        return
        yield

    def monitor_init(self, sim):
        sim.write(self._enabled, self.intf.en)
        return
        yield

    def dataReader(self, sim):
        if self.readPending:
            yield sim.waitOnCombUpdate()
            d = sim.read(self.intf.data)
            self.data.append(d)

            if self.readPending_invalidate:
                self.readPending = False

    def getMonitors(self):
        self.dataReader = OnRisingCallbackLoop(self.clk,
                                               self.dataReader,
                                               self.getEnable)
        return ([self.monitor_init] +
                super(FifoReaderAgent, self).getMonitors() +
                [self.dataReader])

    def monitor(self, sim):
        intf = self.intf
        r = sim.read

        if self.notReset(sim):
            # speculative en set
            yield sim.waitOnCombUpdate()
            wait = r(intf.wait)
            assert wait.vldMask, (intf, sim.now)
            rd = not wait.val
            sim.write(rd, intf.en)

        else:
            sim.write(0, intf.en)
            rd = False

        self.readPending = rd

    def getDrivers(self):
        self.dataWriter = OnRisingCallbackLoop(self.clk,
                                               self.dataWriter,
                                               self.getEnable)
        return ([self.driver_init] +
                super(FifoReaderAgent, self).getDrivers() +
                [self.dataWriter])

    def dataWriter(self, sim):
        # delay data litle bit to have nicer wave
        # otherwise wirte happens before next clk period
        # and it means in 0 time and we will not be able to see it in wave
        yield sim.wait(DEFAULT_CLOCK / 10)
        sim.write(self.lastData, self.intf.data)
        if self.lastData_invalidate:
            self.lastData = None

    def driver(self, sim):
        # now we are before clock event
        # * set wait signal
        # * set last data (done in separate process)
        # * if en == 1, pop next data for next clk
        intf = self.intf
        w = sim.write
        rst_n = self.notReset(sim)
        # speculative write
        if rst_n and self.data:
            wait = 0
        else:
            wait = 1

        w(wait, intf.wait)

        if rst_n:
            yield sim.waitOnCombUpdate()
            # wait for potential update of en
            yield sim.waitOnCombUpdate()
            # check if write can be performed and if it possible do real write

            en = sim.read(intf.en)
            assert en.vldMask, (intf, sim.now)
            if en.val:
                assert self.data, ("underflow", intf, sim.now)
                self.lastData = self.data.popleft()


class FifoWriterAgent(SyncAgentBase):
    """
    Simulation agent for FifoWriter interface
    """

    def __init__(self, intf, allowNoReset=False):
        super(FifoWriterAgent, self).__init__(intf, allowNoReset=allowNoReset)
        self.data = deque()

    def driver_init(self, sim):
        sim.write(self._enabled, self.intf.en)
        return
        yield

    def monitor_init(self, sim):
        sim.write(not self._enabled, self.intf.wait)
        return
        yield

    def setEnable_asDriver(self, en, sim):
        SyncAgentBase.setEnable_asDriver(self, en, sim)
        sim.write(en, self.intf.en)

    def setEnable_asMonitor(self, en, sim):
        SyncAgentBase.setEnable_asMonitor(self, en, sim)
        sim.write(not en, self.intf.wait)

    def monitor(self, sim):
        # set wait signal
        # if en == 1 take data
        intf = self.intf
        sim.write(0, intf.wait)

        yield sim.waitOnCombUpdate()
        # wait for potential update of en
        yield sim.waitOnCombUpdate()

        en = sim.read(intf.en)
        assert en.vldMask
        if en.val:
            yield sim.wait(DEFAULT_CLOCK / 10)
            self.data.append(sim.read(intf.data))

    def driver(self, sim):
        # if wait == 0 set en=1 and set data
        intf = self.intf
        w = sim.write

        if self.notReset(sim) and self.data:
            yield sim.waitOnCombUpdate()

            wait = sim.read(intf.wait)
            assert wait.vldMask
            if not wait.val:
                d = self.data.popleft()
                w(d, intf.data)
                w(1, intf.en)
                return

        w(None, intf.data)
        w(0, intf.en)

    def getDrivers(self):
        return SyncAgentBase.getDrivers(self) + [self.driver_init]

    def getMonitors(self):
        return SyncAgentBase.getMonitors(self) + [self.monitor_init]



from collections import OrderedDict

from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.struct import HStructField
from hwt.hdl.value import Value
from hwt.serializer.generic.indent import getIndent


protectedNames = {"clone", "staticEval",
                  "fromPy", "_dtype", "_usedField", "_val"}


class UnionValBase(Value):
    """
    Base class for values for union types.
    Every union type has it's own value class derived from this.

    :ivar _dtype: union type of this value
    :ivar __usedField: member which is actually used to represent value
    :ivar __val: value for __usedField
    """
    __slots__ = ["_dtype", "_val", "_usedField"]

    def __init__(self, val, typeObj):
        """
        :param val: None or tuple (member name, member value)
        :param typeObj: instance of HUnion HdlType for this value
        """
        self._dtype = typeObj
        if val is not None:
            memberName, v = val
        else:
            memberName = next(iter((typeObj.fields.keys())))
            v = None

        f = self._dtype.fields[memberName]
        if not isinstance(v, Value):
            v = f.dtype.fromPy(v)
        else:
            v._auto_cast(f.dtype)
        self._val = v
        self._usedField = f

    @classmethod
    def fromPy(cls, val, typeObj, vldMask=None):
        """
        :param val: None or tuple (member name, member value)
        :param typeObj: instance of HUnion HdlType for this value
        :param vldMask: if is None validity is resolved from val
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        if vldMask == 0:
            val = None
        return cls(val, typeObj)

    def __repr__(self, indent=0):
        buff = ["{"]
        indentOfFields = getIndent(indent + 1)

        for f in self._dtype.fields.values():
            if f.name is not None:
                val = getattr(self, f.name)
                try:
                    v = val.__repr__(indent=indent + 1)
                except TypeError:
                    v = repr(val)

                buff.append("%s%s: %s" % (indentOfFields, f.name, v))
        buff.append(getIndent(indent) + "}")
        return ("\n").join(buff)


class HUnionMemberHandler(object):

    def __init__(self, field):
        self.field = field

    def set(self, parent, v):
        f = parent._dtype.fields[self.field.name]
        if not isinstance(v, Value):
            v = f.dtype.fromPy(v)
        else:
            v._auto_cast(f.dtype)

        parent._val = v
        parent._usedField = f

    def get(self, parent):
        name = self.field.name
        v = parent._val
        if parent._usedField.name == name:
            return v
        else:
            f = parent._dtype.fields[name]
            v = v._reinterpret_cast(f.dtype)
            parent._val = v
            parent._usedField = f
            return v


class HUnion(HdlType):
    """
    HDL union type (same data multiple representations)

    :ivar fields: read only OrderedDict {key:StructField} for each
        member in this union
    :ivar name: name of this type
    :ivar __bit_length_val: precalculated bit_length of this type
    """

    def __init__(self, *template, name=None):
        """
        :param template: list of tuples (type, name) or HStructField objects
            name can be None (= padding)
        :param name: optional name used for debugging purposes
        """
        self.fields = OrderedDict()
        self.name = name
        bit_length = None

        class UnionVal(UnionValBase):
            pass

        for f in template:
            try:
                field = HStructField(*f)
            except TypeError:
                field = f
            if not isinstance(field, HStructField):
                raise TypeError(
                    "Template for struct field %s is not"
                    " in valid format" % repr(f))

            assert field.name is not None
            self.fields[field.name] = field

            t = field.dtype
            if bit_length is None:
                bit_length = t.bit_length()
            else:
                _bit_length = t.bit_length()
                if _bit_length != bit_length:
                    raise TypeError(
                        field.name, " has different size than others")

            memberHandler = HUnionMemberHandler(field)
            p = property(fget=memberHandler.get, fset=memberHandler.set)
            setattr(UnionVal, field.name, p)

        self.fields = self.fields
        self.__bit_length_val = bit_length

        usedNames = set(self.fields.keys())
        assert not protectedNames.intersection(
            usedNames), protectedNames.intersection(usedNames)

        if name is not None:
            UnionVal.__name__ = name + "Val"

        self.valueCls = UnionVal

    def bit_length(self):
        bl = self.__bit_length_val
        if bl is None:
            raise TypeError("Can not request bit_lenght on type"
                            " which has not fixed size")
        else:
            return self.__bit_length_val

    def getValueCls(self):
        return self.valueCls

    def __fields__eq__(self, other):
        if len(self.fields) != len(other.fields):
            return False

        for k, sf in self.fields.items():
            try:
                of = other.fields[k]
            except KeyError:
                return False

            if (sf.dtype != of.dtype or
                    sf.meta != of.meta):
                return False

        return True

    def __eq__(self, other):
        return self is other or (
            type(self) is type(other) and
            self.bit_length() == other.bit_length() and
            self.__fields__eq__(other))

    def __hash__(self):
        return hash(id(self))

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not None is used as a additional
            information about on which address this type is stored
            (used only by HStruct)
        :param expandStructs: expand HStructTypes (used by HStruct and HArray)
        """
        if self.name:
            name = self.name + " "
        else:
            name = ""

        myIndent = getIndent(indent)
        childIndent = getIndent(indent + 1)
        header = "%sunion %s{" % (myIndent, name)

        buff = [header, ]
        for f in self.fields.values():
            if f.name is None:
                buff.append("%s//%r empty space" % (childIndent, f.dtype))
            else:
                buff.append("%s %s" % (
                                f.dtype.__repr__(indent=indent + 1,
                                                 withAddr=withAddr,
                                                 expandStructs=expandStructs),
                            f.name))

        buff.append("%s}" % (myIndent))
        return "\n".join(buff)




def vldMaskForAnd(a, b):
    # (val, vld)
    # (0, 0) & (0, 0) -> (0, 0)
    # (0, 1) & (0, 0) -> (0, 1)
    # (0, 0) & (0, 1) -> (0, 1)
    # (1, 1) & (0, 0) -> (0, 0)

    a_vld = (a.vldMask & ~a.val)
    b_vld = (b.vldMask & ~b.val)
    vld = (a.vldMask & b.vldMask) | a_vld | b_vld
    return vld


def vldMaskForOr(a, b):
    a_vld = (a.vldMask & a.val)
    b_vld = (b.vldMask & b.val)
    vld = (a.vldMask & b.vldMask) | a_vld | b_vld
    return vld


def vldMaskForXor(a, b):
    return a.vldMask & b.vldMask



from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.boolVal import HBoolVal
from hwt.hdl.types.defs import BOOL
from hwt.hdl.types.hdlType import default_auto_cast_fn
from hwt.hdl.value import Value


def cast_integer(self, sigOrVal, toType):
    isVal = isinstance(sigOrVal, Value)
    if toType == BOOL:
        if isVal:
            v = int(bool(sigOrVal.val))
            return HBoolVal(v, BOOL,
                            sigOrVal.vldMask,
                            sigOrVal.updateTime)

    elif isinstance(toType, Bits):
        if isVal:
            _v = sigOrVal.val
            w = toType.bit_length()
            assert _v.bit_length() <= w,\
                "%d can not fit into %d bits" % (_v, w)
            v = toType.fromPy(_v)

            v.updateTime = sigOrVal.updateTime

            v._dtype = toType
            if not sigOrVal.vldMask:
                v.vldMask = 0
            return v
        else:
            return Operator.withRes(AllOps.IntToBits, [sigOrVal], toType)

    return default_auto_cast_fn(self, sigOrVal, toType)



from hwt.hdl.assignment import Assignment
from hwt.hdl.operator import Operator
from hwt.hdl.statements import IfContainer, SwitchContainer, \
    WhileContainer, WaitStm
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.sliceVal import SliceVal
from hwt.hdl.variables import SignalItem
from hwt.pyUtils.arrayQuery import arr_any
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.generic.indent import getIndent
from hwt.serializer.verilog.utils import verilogTypeOfSig, SIGNAL_TYPE


class VerilogSerializer_statements():
    @classmethod
    def Assignment(cls, a, ctx):
        dst = a.dst
        assert isinstance(dst, SignalItem)

        def valAsHdl(v):
            return cls.Value(v, ctx)

        dstSignalType = verilogTypeOfSig(dst)

        assert not dst.virtualOnly
        if dstSignalType is SIGNAL_TYPE.REG:
            prefix = ""
            symbol = "<="
        else:
            prefix = "assign "
            symbol = "="

        if a.indexes is not None:
            for i in a.indexes:
                if isinstance(i, SliceVal):
                    i = i.clone()
                    i.val = (i.val[0], i.val[1])
                dst = dst[i]

        indent_str = getIndent(ctx.indent)
        dstStr = cls.asHdl(dst, ctx)
        firstPartOfStr = "%s%s%s" % (indent_str, prefix, dstStr)
        src_t = a.src._dtype
        dst_t = dst._dtype

        if dst_t == src_t \
            or (isinstance(src_t, Bits)
                and isinstance(dst_t, Bits)
                and src_t.bit_length() == dst_t.bit_length() == 1):
            return "%s %s %s;" % (firstPartOfStr, symbol, valAsHdl(a.src))
        else:
            raise SerializerException("%s %s %s is not valid assignment\n"
                                      " because types are different (%r; %r) "
                                      % (dstStr, symbol, valAsHdl(a.src),
                                         dst._dtype, a.src._dtype))

    @classmethod
    def HWProcess(cls, proc, ctx):
        """
        Serialize HWProcess objects
        """
        body = proc.statements
        extraVars = []
        extraVarsSerialized = []

        hasToBeVhdlProcess = extraVars or\
            arr_any(body,
                    lambda x: isinstance(x,
                                         (IfContainer,
                                          SwitchContainer,
                                          WhileContainer,
                                          WaitStm)) or
                    (isinstance(x, Assignment) and
                     x.indexes))

        anyIsEventDependnt = arr_any(
            proc.sensitivityList, lambda s: isinstance(s, Operator))
        sensitivityList = sorted(
            map(lambda s: cls.sensitivityListItem(s, ctx,
                                                  anyIsEventDependnt),
                proc.sensitivityList))

        if hasToBeVhdlProcess:
            childCtx = ctx.withIndent()
        else:
            childCtx = ctx

        def createTmpVarFn(suggestedName, dtype):
            # [TODO] it is better to use RtlSignal
            s = SignalItem(None, dtype, virtualOnly=True)
            s.name = childCtx.scope.checkedName(suggestedName, s)
            s.hidden = False
            serializedS = cls.SignalItem(s, childCtx, declaration=True)
            extraVars.append(s)
            extraVarsSerialized.append(serializedS)
            return s

        childCtx.createTmpVarFn = createTmpVarFn

        statemets = [cls.asHdl(s, childCtx) for s in body]

        if hasToBeVhdlProcess:
            proc.name = ctx.scope.checkedName(proc.name, proc)

        extraVarsInit = []
        for s in extraVars:
            a = Assignment(s.defaultVal, s, virtualOnly=True)
            extraVarsInit.append(cls.Assignment(a, childCtx))

        return cls.processTmpl.render(
            indent=getIndent(ctx.indent),
            name=proc.name,
            hasToBeVhdlProcess=hasToBeVhdlProcess,
            extraVars=extraVarsSerialized,
            sensitivityList=" or ".join(sensitivityList),
            statements=extraVarsInit + statemets
        )



from hwt.hdl.value import Value, areValues
from hwt.serializer.generic.indent import getIndent


class StructValBase(Value):
    """
    Base class for values for structure types.
    Every structure type has it's own value class derived from this.
    """
    __slots__ = []

    def __init__(self, val, typeObj, skipCheck=False):
        """
        :param val: None or dict {field name: field value}
        :param typeObj: instance of String HdlType
        :param skipCheck: flag to skip field name consystency in val
        """
        self._dtype = typeObj
        self.updateTime = -1
        if not skipCheck and val is not None:
            assert set(self.__slots__).issuperset(set(val.keys())), \
                set(val.keys()).difference(set(self.__slots__))

        for f in self._dtype.fields:
            if f.name is None:
                continue
            if val is None:
                v = None
            else:
                v = val.get(f.name, None)

            if not isinstance(v, Value):
                v = f.dtype.fromPy(v)

            setattr(self, f.name, v)

    def clone(self):
        d = {}
        for f in self._dtype.fields:
            if f.name is None:
                continue

            v = getattr(self, f.name).clone()
            d[f.name] = v

        return self.__class__(d, self._dtype, skipCheck=True)

    @classmethod
    def fromPy(cls, val, typeObj, vldMask=None):
        """
        :param val: None or dict {field name: field value}
        :param typeObj: instance of String HdlType
        :param vldMask: if is None validity is resolved from val
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        if vldMask == 0:
            val = None
        return cls(val, typeObj)

    def __eq__(self, other):
        if areValues(self, other):
            if self._dtype == other._dtype:
                for f in self._dtype.fields:
                    isPadding = f.name is None
                    if not isPadding:
                        sf = getattr(self, f.name)
                        of = getattr(other, f.name)
                        if not (sf == of):
                            return False
                return True
            else:
                return False
        else:
            return super(Value, self).__eq__(other)

    def __repr__(self, indent=0):
        buff = ["{"]
        indentOfFields = getIndent(indent + 1)

        for f in self._dtype.fields:
            if f.name is not None:
                val = getattr(self, f.name)
                try:
                    v = val.__repr__(indent=indent + 1)
                except TypeError:
                    v = repr(val)

                buff.append("%s%s: %s" % (indentOfFields, f.name, v))
        buff.append(getIndent(indent) + "}")
        return ("\n").join(buff)



from hwt.hdl.types.integer import Integer


class SimInteger(Integer):
    @classmethod
    def getConvertFn(cls):
        raise TypeError("There should not be any need to dynamically convert"
                        "Integer in simulation")


SIM_INT = SimInteger()
# create hdl integer value (for example integer value in vhdl)


def simHInt(val):
    return SIM_INT.fromPy(val)



from hwt.code import Concat
from hwt.hdl.typeShortcuts import vec
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import default_reinterpret_cast_fn, HdlType
from hwt.hdl.value import Value
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


def hstruct_reinterpret_to_bits(self, sigOrVal, toType: HdlType):
    assert toType.bit_length() == self.bit_length()
    parts = []
    for f in self.fields:
        if f.name is None:
            width = f.bit_length()
            part = vec(None, width)
        else:
            part = getattr(sigOrVal, f.name)
            if not isinstance(part, (Value, RtlSignalBase)):
                part = f.dtype.fromPy(part)

        parts.append(part)

    return Concat(*reversed(parts))


def hstruct_reinterpret(self, sigOrVal, toType: HdlType):
    if isinstance(toType, Bits):
        return hstruct_reinterpret_to_bits(self, sigOrVal, toType)
    else:
        return default_reinterpret_cast_fn(self, sigOrVal, toType)




class HdlSimConfig():
    """
    Container of configuration of hdl simulator
    """
    def __init__(self):
        # set to None to prevent redundant calls
        self.beforeSim = None
        self.logChange = None
        self.logPropagation = None
        self.logApplyingValues = None

    def beforeSim(self, simulator, synthesisedUnit):
        """
        called beforee preparing of simulation
        """

    def logChange(self, nowTime, sig, nextVal):
        """
        Log change of value for signal
        """

    def logPropagation(self, simulator, signal, proceses):
        """
        Log value propagation over netlist
        """

    def logApplyingValues(self, simulator, values):
        """
        Log simulator value quantum applied
        """
